from piccolo.columns.column_types import Integer, Varchar
from piccolo.table import Table
from tests.base import TableTest


class Concert(Table):
    """
    ``order`` is a problematic name, as it clashes with a reserved SQL keyword:

    https://www.postgresql.org/docs/current/sql-keywords-appendix.html

    """

    name = Varchar()
    order = Integer()


class TestReservedColumnNames(TableTest):
    """
    Make sure the table works as expected, even though it has a problematic
    column name.
    """

    tables = [Concert]

    def test_common_operations(self):
        # Save / Insert
        concert = Concert(name="Royal Albert Hall", order=1)
        concert.save().run_sync()
        self.assertEqual(
            Concert.select(Concert.order).run_sync(),
            [{"order": 1}],
        )

        # Save / Update
        concert.order = 2
        concert.save().run_sync()
        self.assertEqual(
            Concert.select(Concert.order).run_sync(),
            [{"order": 2}],
        )

        # Update
        Concert.update({Concert.order: 3}, force=True).run_sync()
        self.assertEqual(
            Concert.select(Concert.order).run_sync(),
            [{"order": 3}],
        )

        # Delete
        Concert.delete().where(Concert.order == 3).run_sync()
        self.assertEqual(
            Concert.select(Concert.order).run_sync(),
            [],
        )



import datetime
from unittest import TestCase

import pytest

from piccolo.columns.column_types import (
    Array,
    BigInt,
    Date,
    Integer,
    Time,
    Timestamp,
    Timestamptz,
)
from piccolo.querystring import QueryString
from piccolo.table import Table
from tests.base import TableTest, engines_only, engines_skip, sqlite_only


class MyTable(Table):
    value = Array(base_column=Integer())


class TestArrayDefault(TestCase):
    def test_array_default(self):
        """
        We use ``ListProxy`` instead of ``list`` as a default, because of
        issues with Sphinx's autodoc. Make sure it's correctly converted to a
        plain ``list`` in ``Array.__init__``.
        """
        column = Array(base_column=Integer())
        self.assertTrue(column.default is list)


class TestArray(TableTest):
    """
    Make sure an Array column can be created, and works correctly.
    """

    tables = [MyTable]

    @pytest.mark.cockroach_array_slow
    def test_storage(self):
        """
        Make sure data can be stored and retrieved.

        In CockroachDB <= v22.2.0 we had this error:

        * https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg

        In newer CockroachDB versions, it runs but is very slow:

        * https://github.com/piccolo-orm/piccolo/issues/1005

        """  # noqa: E501
        MyTable(value=[1, 2, 3]).save().run_sync()

        row = MyTable.objects().first().run_sync()
        assert row is not None
        self.assertEqual(row.value, [1, 2, 3])

    @engines_skip("sqlite")
    @pytest.mark.cockroach_array_slow
    def test_index(self):
        """
        Indexes should allow individual array elements to be queried.

        In CockroachDB <= v22.2.0 we had this error:

        * https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg

        In newer CockroachDB versions, it runs but is very slow:

        * https://github.com/piccolo-orm/piccolo/issues/1005

        """  # noqa: E501
        MyTable(value=[1, 2, 3]).save().run_sync()

        self.assertEqual(
            MyTable.select(MyTable.value[0]).first().run_sync(), {"value": 1}
        )

    @engines_skip("sqlite")
    @pytest.mark.cockroach_array_slow
    def test_all(self):
        """
        Make sure rows can be retrieved where all items in an array match a
        given value.

        In CockroachDB <= v22.2.0 we had this error:

        * https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg

        In newer CockroachDB versions, it runs but is very slow:

        * https://github.com/piccolo-orm/piccolo/issues/1005

        """  # noqa: E501
        MyTable(value=[1, 1, 1]).save().run_sync()

        # We have to explicitly specify the type, so CockroachDB works.
        self.assertEqual(
            MyTable.select(MyTable.value)
            .where(MyTable.value.all(QueryString("{}::INTEGER", 1)))
            .first()
            .run_sync(),
            {"value": [1, 1, 1]},
        )

        # We have to explicitly specify the type, so CockroachDB works.
        self.assertEqual(
            MyTable.select(MyTable.value)
            .where(MyTable.value.all(QueryString("{}::INTEGER", 0)))
            .first()
            .run_sync(),
            None,
        )

    @engines_skip("sqlite")
    @pytest.mark.cockroach_array_slow
    def test_any(self):
        """
        Make sure rows can be retrieved where any items in an array match a
        given value.

        In CockroachDB <= v22.2.0 we had this error:

        * https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg

        In newer CockroachDB versions, it runs but is very slow:

        * https://github.com/piccolo-orm/piccolo/issues/1005

        """  # noqa: E501

        MyTable(value=[1, 2, 3]).save().run_sync()

        # We have to explicitly specify the type, so CockroachDB works.
        self.assertEqual(
            MyTable.select(MyTable.value)
            .where(MyTable.value.any(QueryString("{}::INTEGER", 1)))
            .first()
            .run_sync(),
            {"value": [1, 2, 3]},
        )

        # We have to explicitly specify the type, so CockroachDB works.
        self.assertEqual(
            MyTable.select(MyTable.value)
            .where(MyTable.value.any(QueryString("{}::INTEGER", 0)))
            .first()
            .run_sync(),
            None,
        )

    @engines_skip("sqlite")
    @pytest.mark.cockroach_array_slow
    def test_not_any(self):
        """
        Make sure rows can be retrieved where the array doesn't contain a
        certain value.

        In CockroachDB <= v22.2.0 we had this error:

        * https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg

        In newer CockroachDB versions, it runs but is very slow:

        * https://github.com/piccolo-orm/piccolo/issues/1005

        """  # noqa: E501

        MyTable(value=[1, 2, 3]).save().run_sync()
        MyTable(value=[4, 5, 6]).save().run_sync()

        # We have to explicitly specify the type, so CockroachDB works.
        self.assertEqual(
            MyTable.select(MyTable.value)
            .where(MyTable.value.not_any(QueryString("{}::INTEGER", 4)))
            .run_sync(),
            [{"value": [1, 2, 3]}],
        )

    @engines_skip("sqlite")
    @pytest.mark.cockroach_array_slow
    def test_cat(self):
        """
        Make sure values can be appended to an array.

        In CockroachDB <= v22.2.0 we had this error:

        * https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg

        In newer CockroachDB versions, it runs but is very slow:

        * https://github.com/piccolo-orm/piccolo/issues/1005

        """  # noqa: E501
        MyTable(value=[1, 1, 1]).save().run_sync()

        MyTable.update(
            {MyTable.value: MyTable.value.cat([2])}, force=True
        ).run_sync()

        self.assertEqual(
            MyTable.select(MyTable.value).run_sync(),
            [{"value": [1, 1, 1, 2]}],
        )

        # Try plus symbol

        MyTable.update(
            {MyTable.value: MyTable.value + [3]}, force=True
        ).run_sync()

        self.assertEqual(
            MyTable.select(MyTable.value).run_sync(),
            [{"value": [1, 1, 1, 2, 3]}],
        )

        # Make sure non-list values work

        MyTable.update(
            {MyTable.value: MyTable.value + 4}, force=True
        ).run_sync()

        self.assertEqual(
            MyTable.select(MyTable.value).run_sync(),
            [{"value": [1, 1, 1, 2, 3, 4]}],
        )

    @sqlite_only
    def test_cat_sqlite(self):
        """
        If using SQLite then an exception should be raised currently.
        """
        with self.assertRaises(ValueError) as manager:
            MyTable.value.cat([2])

        self.assertEqual(
            str(manager.exception),
            "Only Postgres and Cockroach support array appending.",
        )


###############################################################################
# Date and time arrays


class DateTimeArrayTable(Table):
    date = Array(Date())
    time = Array(Time())
    timestamp = Array(Timestamp())
    timestamptz = Array(Timestamptz())
    date_nullable = Array(Date(), null=True)
    time_nullable = Array(Time(), null=True)
    timestamp_nullable = Array(Timestamp(), null=True)
    timestamptz_nullable = Array(Timestamptz(), null=True)


class TestDateTimeArray(TestCase):
    """
    Make sure that data can be stored and retrieved when using arrays of
    date / time / timestamp.

    We have to serialise / deserialise it in a special way in SQLite, hence
    the tests.

    """

    def setUp(self):
        DateTimeArrayTable.create_table().run_sync()

    def tearDown(self):
        DateTimeArrayTable.alter().drop_table().run_sync()

    @engines_only("postgres", "sqlite")
    def test_storage(self):
        test_date = datetime.date(year=2024, month=1, day=1)
        test_time = datetime.time(hour=12, minute=0)
        test_timestamp = datetime.datetime(
            year=2024, month=1, day=1, hour=12, minute=0
        )
        test_timestamptz = datetime.datetime(
            year=2024,
            month=1,
            day=1,
            hour=12,
            minute=0,
            tzinfo=datetime.timezone.utc,
        )

        DateTimeArrayTable(
            {
                DateTimeArrayTable.date: [test_date],
                DateTimeArrayTable.time: [test_time],
                DateTimeArrayTable.timestamp: [test_timestamp],
                DateTimeArrayTable.timestamptz: [test_timestamptz],
                DateTimeArrayTable.date_nullable: None,
                DateTimeArrayTable.time_nullable: None,
                DateTimeArrayTable.timestamp_nullable: None,
                DateTimeArrayTable.timestamptz_nullable: None,
            }
        ).save().run_sync()

        row = DateTimeArrayTable.objects().first().run_sync()
        assert row is not None

        self.assertListEqual(row.date, [test_date])
        self.assertListEqual(row.time, [test_time])
        self.assertListEqual(row.timestamp, [test_timestamp])
        self.assertListEqual(row.timestamptz, [test_timestamptz])

        self.assertIsNone(row.date_nullable)
        self.assertIsNone(row.time_nullable)
        self.assertIsNone(row.timestamp_nullable)
        self.assertIsNone(row.timestamptz_nullable)


###############################################################################
# Nested arrays


class NestedArrayTable(Table):
    value = Array(base_column=Array(base_column=BigInt()))


class TestNestedArray(TestCase):
    """
    Make sure that tables with nested arrays can be created, and work
    correctly.

    Related to this bug, with nested array columns containing ``BigInt``:

    https://github.com/piccolo-orm/piccolo/issues/936

    """

    def setUp(self):
        NestedArrayTable.create_table().run_sync()

    def tearDown(self):
        NestedArrayTable.alter().drop_table().run_sync()

    @engines_only("postgres", "sqlite")
    def test_storage(self):
        """
        Make sure data can be stored and retrieved.

        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg

        """  # noqa: E501
        NestedArrayTable(value=[[1, 2, 3], [4, 5, 6]]).save().run_sync()

        row = NestedArrayTable.objects().first().run_sync()
        assert row is not None
        self.assertEqual(row.value, [[1, 2, 3], [4, 5, 6]])


class TestGetDimensions(TestCase):
    def test_get_dimensions(self):
        """
        Make sure that `_get_dimensions` returns the correct value.
        """
        self.assertEqual(Array(Integer())._get_dimensions(), 1)
        self.assertEqual(Array(Array(Integer()))._get_dimensions(), 2)
        self.assertEqual(Array(Array(Array(Integer())))._get_dimensions(), 3)


class TestGetInnerValueType(TestCase):
    def test_get_inner_value_type(self):
        """
        Make sure that `_get_inner_value_type` returns the correct base type.
        """
        self.assertEqual(Array(Integer())._get_inner_value_type(), int)
        self.assertEqual(Array(Array(Integer()))._get_inner_value_type(), int)
        self.assertEqual(
            Array(Array(Array(Integer())))._get_inner_value_type(), int
        )



from unittest import TestCase

from piccolo.columns import ForeignKey, Varchar
from piccolo.table import Table, create_db_tables_sync, drop_db_tables_sync


class Manager(Table):
    theName = Varchar()


class Band(Table):
    theName = Varchar()
    theManager = ForeignKey(Manager)


class TestCamelCase(TestCase):
    def setUp(self):
        create_db_tables_sync(Manager, Band)

    def tearDown(self):
        drop_db_tables_sync(Manager, Band)

    def test_queries(self):
        """
        Make sure that basic queries work when the columns use camelCase.
        """
        manager_names = ("Guido", "Maz", "Graydon")
        band_names = ("Pythonistas", "Rubyists", "Rustaceans")

        # Test create
        for manager_name, band_name in zip(manager_names, band_names):
            manager = Manager.objects().create(theName=manager_name).run_sync()
            Band.objects().create(
                theName=band_name, theManager=manager
            ).run_sync()

        # Test select, with joins
        response = (
            Band.select(
                Band.theName,
                Band.theManager.theName.as_alias("theManagerName"),
            )
            .order_by(Band.theName)
            .run_sync()
        )
        self.assertListEqual(
            response,
            [
                {"theName": "Pythonistas", "theManagerName": "Guido"},
                {"theName": "Rubyists", "theManagerName": "Maz"},
                {"theName": "Rustaceans", "theManagerName": "Graydon"},
            ],
        )

        # Test delete
        Band.delete().where(Band.theName == "Rubyists").run_sync()

        # Test exists
        self.assertFalse(
            Band.exists().where(Band.theName == "Rubyists").run_sync()
        )



import pytest

from piccolo.query.functions.string import Concat, Upper
from tests.base import engine_version_lt, is_running_sqlite
from tests.example_apps.music.tables import Band

from .base import BandTest


class TestUpper(BandTest):

    def test_column(self):
        """
        Make sure we can uppercase a column's value.
        """
        response = Band.select(Upper(Band.name)).run_sync()
        self.assertListEqual(response, [{"upper": "PYTHONISTAS"}])

    def test_alias(self):
        response = Band.select(Upper(Band.name, alias="name")).run_sync()
        self.assertListEqual(response, [{"name": "PYTHONISTAS"}])

    def test_joined_column(self):
        """
        Make sure we can uppercase a column's value from a joined table.
        """
        response = Band.select(Upper(Band.manager._.name)).run_sync()
        self.assertListEqual(response, [{"upper": "GUIDO"}])


@pytest.mark.skipif(
    is_running_sqlite() and engine_version_lt(3.44),
    reason="SQLite version not supported",
)
class TestConcat(BandTest):

    def test_column_and_string(self):
        response = Band.select(
            Concat(Band.name, "!!!", alias="name")
        ).run_sync()
        self.assertListEqual(response, [{"name": "Pythonistas!!!"}])

    def test_column_and_column(self):
        response = Band.select(
            Concat(Band.name, Band.popularity, alias="name")
        ).run_sync()
        self.assertListEqual(response, [{"name": "Pythonistas1000"}])

    def test_join(self):
        response = Band.select(
            Concat(Band.name, "-", Band.manager._.name, alias="name")
        ).run_sync()
        self.assertListEqual(response, [{"name": "Pythonistas-Guido"}])

    def test_min_args(self):
        with self.assertRaises(ValueError):
            Concat()



import uuid

from piccolo.columns.column_types import UUID
from piccolo.table import Table
from tests.base import TableTest


class MyTable(Table):
    uuid = UUID()


class TestUUID(TableTest):
    tables = [MyTable]

    def test_return_type(self):
        row = MyTable()
        row.save().run_sync()

        self.assertIsInstance(row.uuid, uuid.UUID)



import typing as t

from tests.base import DBTestCase, TableTest
from tests.example_apps.music.tables import (
    Band,
    Concert,
    Manager,
    RecordingStudio,
    Venue,
)


class TestRefresh(DBTestCase):
    def setUp(self):
        super().setUp()
        self.insert_rows()

    def test_refresh(self) -> None:
        """
        Make sure ``refresh`` works, with no columns specified.
        """
        # Fetch an instance from the database.
        band = Band.objects().get(Band.name == "Pythonistas").run_sync()
        assert band is not None
        initial_data = band.to_dict()

        # Modify the data in the database.
        Band.update(
            {Band.name: Band.name + "!!!", Band.popularity: 8000}
        ).where(Band.name == "Pythonistas").run_sync()

        # Refresh `band`, and make sure it has the correct data.
        band.refresh().run_sync()

        self.assertEqual(band.name, "Pythonistas!!!")
        self.assertEqual(band.popularity, 8000)
        self.assertEqual(band.id, initial_data["id"])

    def test_refresh_with_prefetch(self) -> None:
        """
        Make sure ``refresh`` works, when the object used prefetch to get
        nested objets (the nested objects should be updated too).
        """
        band = (
            Band.objects(Band.manager)
            .where(Band.name == "Pythonistas")
            .first()
            .run_sync()
        )
        assert band is not None

        # Modify the data in the database.
        Manager.update({Manager.name: "Guido!!!"}).where(
            Manager.name == "Guido"
        ).run_sync()

        # Refresh `band`, and make sure it has the correct data.
        band.refresh().run_sync()

        self.assertEqual(band.manager.name, "Guido!!!")

    def test_refresh_with_prefetch_multiple_layers_deep(self) -> None:
        """
        Make sure ``refresh`` works, when the object used prefetch to get
        nested objets (the nested objects should be updated too).
        """
        band = (
            Band.objects(Band.manager)
            .where(Band.name == "Pythonistas")
            .first()
            .run_sync()
        )
        assert band is not None

        # Modify the data in the database.
        Manager.update({Manager.name: "Guido!!!"}).where(
            Manager.name == "Guido"
        ).run_sync()

        # Refresh `band`, and make sure it has the correct data.
        band.refresh().run_sync()

        self.assertEqual(band.manager.name, "Guido!!!")

    def test_columns(self) -> None:
        """
        Make sure ``refresh`` works, when columns are specified.
        """
        # Fetch an instance from the database.
        band = Band.objects().get(Band.name == "Pythonistas").run_sync()
        assert band is not None
        initial_data = band.to_dict()

        # Modify the data in the database.
        Band.update(
            {Band.name: Band.name + "!!!", Band.popularity: 8000}
        ).where(Band.name == "Pythonistas").run_sync()

        # Refresh `band`, and make sure it has the correct data.
        query = band.refresh(columns=[Band.name])
        self.assertEqual(
            [i._meta.name for i in query._columns],
            ["name"],
        )
        query.run_sync()

        self.assertEqual(band.name, "Pythonistas!!!")
        self.assertEqual(band.popularity, initial_data["popularity"])
        self.assertEqual(band.id, initial_data["id"])

    def test_error_when_not_in_db(self) -> None:
        """
        Make sure we can't refresh an instance which hasn't been saved in the
        database.
        """
        band = Band()

        with self.assertRaises(ValueError) as manager:
            band.refresh().run_sync()

        self.assertEqual(
            "The instance doesn't exist in the database.",
            str(manager.exception),
        )

    def test_error_when_pk_in_none(self) -> None:
        """
        Make sure we can't refresh an instance when the primary key value isn't
        set.
        """
        band = Band.objects().first().run_sync()
        assert band is not None
        band.id = None

        with self.assertRaises(ValueError) as manager:
            band.refresh().run_sync()

        self.assertEqual(
            "The instance's primary key value isn't defined.",
            str(manager.exception),
        )


class TestRefreshWithPrefetch(TableTest):

    tables = [Manager, Band, Concert, Venue]

    def setUp(self):
        super().setUp()

        self.manager = Manager({Manager.name: "Guido"})
        self.manager.save().run_sync()

        self.band = Band(
            {Band.name: "Pythonistas", Band.manager: self.manager}
        )
        self.band.save().run_sync()

        self.concert = Concert({Concert.band_1: self.band})
        self.concert.save().run_sync()

    def test_single_layer(self) -> None:
        """
        Make sure ``refresh`` works, when the object used prefetch to get
        nested objects (the nested objects should be updated too).
        """
        band = (
            Band.objects(Band.manager)
            .where(Band.name == "Pythonistas")
            .first()
            .run_sync()
        )
        assert band is not None

        # Modify the data in the database.
        self.manager.name = "Guido!!!"
        self.manager.save().run_sync()

        # Refresh `band`, and make sure it has the correct data.
        band.refresh().run_sync()
        self.assertEqual(band.manager.name, "Guido!!!")

    def test_multiple_layers(self) -> None:
        """
        Make sure ``refresh`` works when ``prefetch`` was used to fetch objects
        multiple layers deep.
        """
        concert = (
            Concert.objects(Concert.band_1._.manager)
            .where(Concert.band_1._.name == "Pythonistas")
            .first()
            .run_sync()
        )
        assert concert is not None

        # Modify the data in the database.
        self.manager.name = "Guido!!!"
        self.manager.save().run_sync()

        concert.refresh().run_sync()
        self.assertEqual(concert.band_1.manager.name, "Guido!!!")

    def test_updated_foreign_key(self) -> None:
        """
        If a foreign key now references a different row, make sure this
        is refreshed correctly.
        """
        band = (
            Band.objects(Band.manager)
            .where(Band.name == "Pythonistas")
            .first()
            .run_sync()
        )
        assert band is not None

        # Assign a different manager to the band
        new_manager = Manager({Manager.name: "New Manager"})
        new_manager.save().run_sync()
        Band.update({Band.manager: new_manager.id}, force=True).run_sync()

        # Refresh `band`, and make sure it references the new manager.
        band.refresh().run_sync()
        self.assertEqual(band.manager.id, new_manager.id)
        self.assertEqual(band.manager.name, "New Manager")

    def test_foreign_key_set_to_null(self):
        """
        Make sure that if the foreign key was set to null, that ``refresh``
        sets the nested object to ``None``.
        """
        band = (
            Band.objects(Band.manager)
            .where(Band.name == "Pythonistas")
            .first()
            .run_sync()
        )
        assert band is not None

        # Remove the manager from band
        Band.update({Band.manager: None}, force=True).run_sync()

        # Refresh `band`, and make sure the foreign key value is now `None`,
        # instead of a nested object.
        band.refresh().run_sync()
        self.assertIsNone(band.manager)

    def test_exception(self) -> None:
        """
        We don't currently let the user refresh specific fields from nested
        objects - an exception should be raised.
        """
        with self.assertRaises(ValueError):
            self.concert.refresh(columns=[Concert.band_1._.manager]).run_sync()

        # Shouldn't raise an exception:
        self.concert.refresh(columns=[Concert.band_1]).run_sync()


class TestRefreshWithLoadJSON(TableTest):

    tables = [RecordingStudio]

    def setUp(self):
        super().setUp()

        self.recording_studio = RecordingStudio(
            {RecordingStudio.facilities: {"piano": True}}
        )
        self.recording_studio.save().run_sync()

    def test_load_json(self):
        """
        Make sure we can refresh an object, and load the JSON as a Python
        object.
        """
        RecordingStudio.update(
            {RecordingStudio.facilities: {"electric piano": True}},
            force=True,
        ).run_sync()

        # Refresh without load_json:
        self.recording_studio.refresh().run_sync()

        self.assertEqual(
            # Remove the white space, because some versions of Python add
            # whitespace around JSON, and some don't.
            self.recording_studio.facilities.replace(" ", ""),
            '{"electricpiano":true}',
        )

        # Refresh with load_json:
        self.recording_studio.refresh(load_json=True).run_sync()

        self.assertDictEqual(
            t.cast(dict, self.recording_studio.facilities),
            {"electric piano": True},
        )



from piccolo.conf.apps import AppConfig

from .commands.run import run

APP_CONFIG = AppConfig(
    app_name="tester",
    migrations_folder_path="",
    table_classes=[],
    migration_dependencies=[],
    commands=[run],
)



from __future__ import annotations

import typing as t

from piccolo.utils.encoding import JSONDict
from piccolo.utils.sync import run_sync

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns import Column
    from piccolo.table import Table


class Refresh:
    """
    Used to refresh :class:`Table <piccolo.table.Table>` instances with the
    latest data data from the database. Accessible via
    :meth:`refresh <piccolo.table.Table.refresh>`.

    :param instance:
        The instance to refresh.
    :param columns:
        Which columns to refresh - it not specified, then all columns are
        refreshed.
    :param load_json:
        Whether to load ``JSON`` / ``JSONB`` columns as objects, instead of
        just a string.

    """

    def __init__(
        self,
        instance: Table,
        columns: t.Optional[t.Sequence[Column]] = None,
        load_json: bool = False,
    ):
        self.instance = instance

        if columns:
            for column in columns:
                if len(column._meta.call_chain) > 0:
                    raise ValueError(
                        "We can't currently selectively refresh certain "
                        "columns on child objects (e.g. Concert.band_1.name). "
                        "Please just specify top level columns (e.g. "
                        "Concert.band_1), and the entire child object will be "
                        "refreshed."
                    )

        self.columns = columns
        self.load_json = load_json

    @property
    def _columns(self) -> t.Sequence[Column]:
        """
        Works out which columns the user wants to refresh.
        """
        if self.columns:
            return self.columns

        return [
            i for i in self.instance._meta.columns if not i._meta.primary_key
        ]

    def _get_columns(self, instance: Table, columns: t.Sequence[Column]):
        """
        If `prefetch` was used on the object, for example::

            >>> await Band.objects(Band.manager)

        We should also update the prefetched object.

        It works multiple level deep. If we refresh this::

            >>> await Album.objects(Album.band.manager).first()

        It will update the nested `band` object, and also the `manager`
        object.

        """
        from piccolo.columns.column_types import ForeignKey
        from piccolo.table import Table

        select_columns = []

        for column in columns:
            if isinstance(column, ForeignKey) and isinstance(
                (child_instance := getattr(instance, column._meta.name)),
                Table,
            ):
                select_columns.extend(
                    self._get_columns(
                        child_instance,
                        # Fetch all columns (even the primary key, just in
                        # case the foreign key now references a different row).
                        column.all_columns(),
                    )
                )
            else:
                select_columns.append(column)

        return select_columns

    def _update_instance(self, instance: Table, data_dict: t.Dict):
        """
        Update the table instance. It is called recursively, if the instance
        has child instances.
        """
        for key, value in data_dict.items():
            if isinstance(value, dict) and not isinstance(value, JSONDict):
                # If the value is a dict, then it's a child instance.
                if all(i is None for i in value.values()):
                    # If all values in the nested object are None, then we can
                    # safely assume that the object itself is null, as the
                    # primary key value must be null.
                    setattr(instance, key, None)
                else:
                    self._update_instance(getattr(instance, key), value)
            else:
                setattr(instance, key, value)

    async def run(
        self, in_pool: bool = True, node: t.Optional[str] = None
    ) -> Table:
        """
        Run it asynchronously. For example::

            await my_instance.refresh().run()

            # or for convenience:
            await my_instance.refresh()

        Modifies the instance in place, but also returns it as a convenience.

        """
        instance = self.instance

        if not instance._exists_in_db:
            raise ValueError("The instance doesn't exist in the database.")

        pk_column = instance._meta.primary_key

        primary_key_value = getattr(instance, pk_column._meta.name, None)

        if primary_key_value is None:
            raise ValueError("The instance's primary key value isn't defined.")

        columns = self._columns
        if not columns:
            raise ValueError("No columns to fetch.")

        select_columns = self._get_columns(
            instance=self.instance, columns=columns
        )

        data_dict = (
            await instance.__class__.select(*select_columns)
            .where(pk_column == primary_key_value)
            .output(nested=True, load_json=self.load_json)
            .first()
            .run(node=node, in_pool=in_pool)
        )

        if data_dict is None:
            raise ValueError(
                "The object doesn't exist in the database any more."
            )

        self._update_instance(instance=instance, data_dict=data_dict)

        return instance

    def __await__(self):
        """
        If the user doesn't explicity call :meth:`run`, proxy to it as a
        convenience.
        """
        return self.run().__await__()

    def run_sync(self, *args, **kwargs) -> Table:
        """
        Run it synchronously. For example::

            my_instance.refresh().run_sync()

        """
        return run_sync(self.run(*args, **kwargs))



from tests.base import TableTest
from tests.example_apps.music.tables import Band, Manager


class BandTest(TableTest):
    tables = [Band, Manager]

    def setUp(self) -> None:
        super().setUp()

        manager = Manager({Manager.name: "Guido"})
        manager.save().run_sync()

        band = Band(
            {
                Band.name: "Pythonistas",
                Band.manager: manager,
                Band.popularity: 1000,
            }
        )
        band.save().run_sync()



from piccolo.columns.column_types import Real
from piccolo.table import Table
from tests.base import TableTest


class MyTable(Table):
    column_a = Real()


class TestReal(TableTest):
    tables = [MyTable]

    def test_creation(self):
        row = MyTable(column_a=1.23)
        row.save().run_sync()

        _row = MyTable.objects().first().run_sync()
        assert _row is not None
        self.assertEqual(type(_row.column_a), float)
        self.assertAlmostEqual(_row.column_a, 1.23)



from piccolo.columns.column_types import Varchar
from piccolo.table import Table

from ..base import TableTest, engines_only


class MyTable(Table):
    name = Varchar(length=10)


@engines_only("postgres", "cockroach")
class TestVarchar(TableTest):
    """
    SQLite doesn't enforce any constraints on max character length.

    https://www.sqlite.org/faq.html#q9

    Might consider enforcing this at the ORM level instead in the future.
    """

    tables = [MyTable]

    def test_length(self):
        row = MyTable(name="bob")
        row.save().run_sync()

        with self.assertRaises(Exception):
            row.name = "bob123456789"
            row.save().run_sync()



from .base import Engine
from .cockroach import CockroachEngine
from .finder import engine_finder
from .postgres import PostgresEngine
from .sqlite import SQLiteEngine

__all__ = [
    "Engine",
    "PostgresEngine",
    "SQLiteEngine",
    "CockroachEngine",
    "engine_finder",
]



from __future__ import annotations

import typing as t


def make_nested(dictionary: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    """
    Rows are returned from the database as a flat dictionary, with keys such
    as ``'manager.name'`` if the column belongs to a related table.

    This function puts any values from a related table into a sub dictionary.

    .. code-block:: python

        response = await Band.select(Band.name, Band.manager.name)
        >>> print(response)
        [{'name': 'Pythonistas', 'band.name': 'Guido'}]

        >>> make_nested(response[0])
        {'name': 'Pythonistas', 'band': {'name': 'Guido'}}

    """
    output: t.Dict[str, t.Any] = {}

    items = list(dictionary.items())
    items.sort(key=lambda x: x[0])

    for key, value in items:
        path = key.split(".")
        if len(path) == 1:
            output[path[0]] = value
        else:
            # Force the root element to be an empty dictionary, if it's some
            # other value (most likely an integer). This is because there are
            # situations where a query can have `band` and `band.id`.
            # For example:
            # await Band.select(
            #     Band.all_columns(),
            #     Band.manager.all_columns()
            # ).run()
            # In this situation nesting takes precendence.
            root = output.get(path[0], None)
            if isinstance(root, dict):
                dictionary = root
            else:
                dictionary = {}
                output[path[0]] = dictionary

            for path_element in path[1:-1]:
                root = dictionary.setdefault(path_element, {})
                if not isinstance(root, dict):
                    root = {}
                    dictionary[path_element] = root
                dictionary = root

            dictionary[path[-1]] = value

    return output



import typing as t

from piccolo.columns.base import Column
from piccolo.querystring import QueryString


class Cast(QueryString):
    def __init__(
        self,
        identifier: t.Union[Column, QueryString],
        as_type: Column,
        alias: t.Optional[str] = None,
    ):
        """
        Cast a value to a different type. For example::

            >>> from piccolo.query.functions import Cast

            >>> await Concert.select(
            ...     Cast(Concert.starts, Time(), "start_time")
            ... )
            [{"start_time": datetime.time(19, 0)}]

        :param identifier:
            Identifies what is being converted (e.g. a column).
        :param as_type:
            The type to be converted to.

        """
        # Make sure the identifier is a supported type.

        if not isinstance(identifier, (Column, QueryString)):
            raise ValueError(
                "The identifier is an unsupported type - only Column and "
                "QueryString instances are allowed."
            )

        #######################################################################
        # Convert `as_type` to a string which can be used in the query.

        if not isinstance(as_type, Column):
            raise ValueError("The `as_type` value must be a Column instance.")

        # We need to give the column a reference to a table, and hence
        # the database engine, as the column type is sometimes dependent
        # on which database is being used.
        from piccolo.table import Table, create_table_class

        table: t.Optional[t.Type[Table]] = None

        if isinstance(identifier, Column):
            table = identifier._meta.table
        elif isinstance(identifier, QueryString):
            table = (
                identifier.columns[0]._meta.table
                if identifier.columns
                else None
            )

        as_type._meta.table = table or create_table_class("Table")
        as_type_string = as_type.column_type

        #######################################################################
        # Preserve the original alias from the column.

        if isinstance(identifier, Column):
            alias = (
                alias
                or identifier._alias
                or identifier._meta.get_default_alias()
            )

        #######################################################################

        super().__init__(
            f"CAST({{}} AS {as_type_string})",
            identifier,
            alias=alias,
        )


__all__ = ("Cast",)



import asyncio
import json
import typing as t
import unittest

from piccolo.columns import (
    Array,
    Decimal,
    ForeignKey,
    Integer,
    LazyTableReference,
    Numeric,
    Real,
    Timestamp,
    Timestamptz,
    Varchar,
)
from piccolo.table import Table, create_db_tables_sync, drop_db_tables_sync
from piccolo.testing.model_builder import ModelBuilder
from tests.base import engines_skip
from tests.example_apps.music.tables import (
    Band,
    Concert,
    Manager,
    Poster,
    RecordingStudio,
    Shirt,
    Ticket,
    Venue,
)


class TableWithArrayField(Table):
    strings = Array(Varchar(30))
    integers = Array(Integer())
    floats = Array(Real())


class TableWithDecimal(Table):
    numeric = Numeric()
    numeric_with_digits = Numeric(digits=(4, 2))
    decimal = Decimal()
    decimal_with_digits = Decimal(digits=(4, 2))


class BandWithLazyReference(Table):
    manager: ForeignKey["Manager"] = ForeignKey(
        references=LazyTableReference(
            "Manager", module_path="tests.example_apps.music.tables"
        )
    )


class BandWithRecursiveReference(Table):
    manager: ForeignKey["Manager"] = ForeignKey("self")


TABLES = (
    Manager,
    Band,
    Poster,
    RecordingStudio,
    Shirt,
    Venue,
    Concert,
    Ticket,
    TableWithArrayField,
    TableWithDecimal,
    BandWithLazyReference,
    BandWithRecursiveReference,
)


# Cockroach Bug: Can turn ON when resolved: https://github.com/cockroachdb/cockroach/issues/71908  # noqa: E501
@engines_skip("cockroach")
class TestModelBuilder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        create_db_tables_sync(*TABLES)

    @classmethod
    def tearDownClass(cls) -> None:
        drop_db_tables_sync(*TABLES)

    def test_async(self):
        async def build_model(table_class: t.Type[Table]):
            return await ModelBuilder.build(table_class)

        for table_class in TABLES:
            asyncio.run(build_model(table_class))

    def test_sync(self):
        for table_class in TABLES:
            ModelBuilder.build_sync(table_class)

    def test_choices(self):
        shirt = ModelBuilder.build_sync(Shirt)
        queried_shirt = (
            Shirt.objects().where(Shirt.id == shirt.id).first().run_sync()
        )
        assert queried_shirt is not None

        self.assertIn(
            queried_shirt.size,
            ["s", "l", "m"],
        )

    def test_datetime(self):
        """
        Make sure that ``ModelBuilder`` generates timezone aware datetime
        objects for ``Timestamptz`` columns, and timezone naive datetime
        objects for ``Timestamp`` columns.
        """

        class Table1(Table):
            starts = Timestamptz()

        class Table2(Table):
            starts = Timestamp()

        model_1 = ModelBuilder.build_sync(Table1, persist=False)
        assert model_1.starts.tzinfo is not None

        model_2 = ModelBuilder.build_sync(Table2, persist=False)
        assert model_2.starts.tzinfo is None

    def test_foreign_key(self):
        model = ModelBuilder.build_sync(Band, persist=True)

        self.assertTrue(
            Manager.exists().where(Manager.id == model.manager).run_sync()
        )

    def test_lazy_foreign_key(self):
        model = ModelBuilder.build_sync(BandWithLazyReference, persist=True)

        self.assertTrue(
            Manager.exists().where(Manager.id == model.manager).run_sync()
        )

    def test_recursive_foreign_key(self):
        """
        Make sure no infinite loops are created with recursive foreign keys.
        """
        model = ModelBuilder.build_sync(
            BandWithRecursiveReference, persist=True
        )
        # It should be set to None, as this foreign key is nullable.
        self.assertIsNone(model.manager)

    def test_invalid_column(self):
        with self.assertRaises(ValueError):
            ModelBuilder.build_sync(Band, defaults={"X": 1})

    def test_minimal(self):
        band = ModelBuilder.build_sync(Band, minimal=True)

        self.assertTrue(Band.exists().where(Band.id == band.id).run_sync())

    def test_persist_false(self):
        band = ModelBuilder.build_sync(Band, persist=False)

        self.assertFalse(Band.exists().where(Band.id == band.id).run_sync())

    def test_valid_column(self):
        manager = ModelBuilder.build_sync(
            Manager, defaults={Manager.name: "Guido"}
        )

        queried_manager = (
            Manager.objects()
            .where(Manager.id == manager.id)
            .first()
            .run_sync()
        )
        assert queried_manager is not None

        self.assertEqual(queried_manager.name, "Guido")

    def test_valid_column_string(self):
        manager = ModelBuilder.build_sync(Manager, defaults={"name": "Guido"})

        queried_manager = (
            Manager.objects()
            .where(Manager.id == manager.id)
            .first()
            .run_sync()
        )
        assert queried_manager is not None

        self.assertEqual(queried_manager.name, "Guido")

    def test_valid_foreign_key(self):
        manager = ModelBuilder.build_sync(Manager)

        band = ModelBuilder.build_sync(Band, defaults={Band.manager: manager})

        self.assertEqual(manager._meta.primary_key, band.manager)

    def test_valid_foreign_key_string(self):
        manager = ModelBuilder.build_sync(Manager)

        band = ModelBuilder.build_sync(Band, defaults={"manager": manager})

        self.assertEqual(manager._meta.primary_key, band.manager)

    def test_json(self):
        """
        Make sure the generated JSON can be parsed.

        This is important, because we might have queries like this::

            >>> await RecordingStudio.select().output(load_json=True)

        """
        studio = ModelBuilder.build_sync(RecordingStudio)
        self.assertIsInstance(json.loads(studio.facilities), dict)
        self.assertIsInstance(json.loads(studio.facilities_b), dict)

        for facilities in (
            RecordingStudio.select(RecordingStudio.facilities)
            .output(load_json=True, as_list=True)
            .run_sync()
        ):
            self.assertIsInstance(facilities, dict)



"""
Most of the tests for piccolo/columns/reference.py are covered in
piccolo/columns/test_foreignkey.py
"""

from unittest import TestCase

from piccolo.columns import ForeignKey, Varchar
from piccolo.columns.reference import LazyTableReference
from piccolo.table import Table
from tests.base import TableTest


class Band(Table):
    manager: ForeignKey["Manager"] = ForeignKey(
        LazyTableReference("Manager", module_path=__name__)
    )
    name = Varchar()


class Manager(Table):
    name = Varchar()


class TestQueries(TableTest):
    tables = [Band, Manager]

    def setUp(self):
        super().setUp()
        manager = Manager({Manager.name: "Guido"})
        manager.save().run_sync()
        band = Band({Band.name: "Pythonistas", Band.manager: manager})
        band.save().run_sync()

    def test_select(self):
        self.assertListEqual(
            Band.select(Band.name, Band.manager._.name).run_sync(),
            [{"name": "Pythonistas", "manager.name": "Guido"}],
        )


class TestInit(TestCase):
    def test_init(self):
        """
        A ``LazyTableReference`` must be passed either an ``app_name`` or
        ``module_path`` argument.
        """
        with self.assertRaises(ValueError):
            LazyTableReference(table_class_name="Manager")

        with self.assertRaises(ValueError):
            LazyTableReference(
                table_class_name="Manager",
                app_name="music",
                module_path="tests.example_apps.music.tables",
            )

        # Shouldn't raise exceptions:
        LazyTableReference(
            table_class_name="Manager",
            app_name="music",
        )
        LazyTableReference(
            table_class_name="Manager",
            module_path="tests.example_apps.music.tables",
        )


class TestStr(TestCase):
    def test_str(self):
        self.assertEqual(
            LazyTableReference(
                table_class_name="Manager",
                app_name="music",
            ).__str__(),
            "App music.Manager",
        )

        self.assertEqual(
            LazyTableReference(
                table_class_name="Manager",
                module_path="tests.example_apps.music.tables",
            ).__str__(),
            "Module tests.example_apps.music.tables.Manager",
        )



from tests.base import DBTestCase, engine_is
from tests.example_apps.music.tables import Band


class TestRaw(DBTestCase):
    def test_raw_without_args(self):
        self.insert_row()

        response = Band.raw("SELECT * FROM band").run_sync()

        if engine_is("cockroach"):
            self.assertDictEqual(
                response[0],
                {
                    "id": response[0]["id"],
                    "name": "Pythonistas",
                    "manager": response[0]["manager"],
                    "popularity": 1000,
                },
            )
        else:
            self.assertDictEqual(
                response[0],
                {
                    "id": 1,
                    "name": "Pythonistas",
                    "manager": 1,
                    "popularity": 1000,
                },
            )

    def test_raw_with_args(self):
        self.insert_rows()

        response = Band.raw(
            "SELECT * FROM band WHERE name = {}", "Pythonistas"
        ).run_sync()

        self.assertEqual(len(response), 1)

        if engine_is("cockroach"):
            self.assertDictEqual(
                response[0],
                {
                    "id": response[0]["id"],
                    "name": "Pythonistas",
                    "manager": response[0]["manager"],
                    "popularity": 1000,
                },
            )
        else:
            self.assertDictEqual(
                response[0],
                {
                    "id": 1,
                    "name": "Pythonistas",
                    "manager": 1,
                    "popularity": 1000,
                },
            )



from unittest import TestCase
from unittest.mock import AsyncMock, MagicMock, patch

from piccolo.apps.user.commands.list import list_users
from piccolo.apps.user.tables import BaseUser
from piccolo.utils.sync import run_sync


class TestList(TestCase):
    def setUp(self):
        BaseUser.create_table(if_not_exists=True).run_sync()
        self.username = "test_user"
        self.password = "abc123XYZ"
        self.user = BaseUser.create_user_sync(
            username=self.username, password=self.password
        )

    def tearDown(self):
        BaseUser.alter().drop_table().run_sync()

    @patch("piccolo.utils.printing.print")
    def test_list(self, print_mock: MagicMock):
        """
        Make sure the user information is listed, excluding the password.
        """
        run_sync(list_users())

        output = "\n".join(i.args[0] for i in print_mock.call_args_list)

        assert self.username in output
        assert self.password not in output
        assert self.user.password not in output


class TestLimit(TestCase):
    def test_non_positive(self):
        """
        Make sure non-positive `limit` values are rejected.
        """
        for value in (0, -1):
            with self.assertRaises(ValueError):
                run_sync(list_users(page=value))


class TestPage(TestCase):
    def test_non_positive(self):
        """
        Make sure non-positive `page` values are rejected.
        """
        for value in (0, -1):
            with self.assertRaises(ValueError):
                run_sync(list_users(limit=value))


class TestOrder(TestCase):
    @patch("piccolo.apps.user.commands.list.get_users")
    def test_order(self, get_users: AsyncMock):
        """
        Make sure valid column names are accepted.
        """
        get_users.return_value = []
        run_sync(list_users(order_by="email"))

        self.assertDictEqual(
            get_users.call_args.kwargs,
            {
                "order_by": BaseUser.email,
                "ascending": True,
                "limit": 20,
                "page": 1,
            },
        )

    @patch("piccolo.apps.user.commands.list.get_users")
    def test_descending(self, get_users: AsyncMock):
        """
        Make sure a colume name prefixed with '-' works.
        """
        get_users.return_value = []
        run_sync(list_users(order_by="-email"))

        self.assertDictEqual(
            get_users.call_args.kwargs,
            {
                "order_by": BaseUser.email,
                "ascending": False,
                "limit": 20,
                "page": 1,
            },
        )

    def test_unrecognised_column(self):
        """
        Make sure invalid column names are rejected.
        """
        with self.assertRaises(ValueError):
            run_sync(list_users(order_by="abc123"))



from __future__ import annotations

import inspect
import logging
import typing as t
from dataclasses import dataclass, field

from piccolo.apps.migrations.auto.diffable_table import DiffableTable
from piccolo.apps.migrations.auto.operations import (
    AlterColumn,
    ChangeTableSchema,
    DropColumn,
    RenameColumn,
    RenameTable,
)
from piccolo.apps.migrations.auto.serialisation import deserialise_params
from piccolo.columns import Column, column_types
from piccolo.columns.column_types import ForeignKey, Serial
from piccolo.engine import engine_finder
from piccolo.query import Query
from piccolo.query.base import DDL
from piccolo.schema import SchemaDDLBase
from piccolo.table import Table, create_table_class, sort_table_classes
from piccolo.utils.warnings import colored_warning

logger = logging.getLogger(__name__)


@dataclass
class AddColumnClass:
    column: Column
    table_class_name: str
    tablename: str
    schema: t.Optional[str]


@dataclass
class AddColumnCollection:
    add_columns: t.List[AddColumnClass] = field(default_factory=list)

    def append(self, add_column: AddColumnClass):
        self.add_columns.append(add_column)

    def for_table_class_name(
        self, table_class_name: str
    ) -> t.List[AddColumnClass]:
        return [
            i
            for i in self.add_columns
            if i.table_class_name == table_class_name
        ]

    def columns_for_table_class_name(
        self, table_class_name: str
    ) -> t.List[Column]:
        return [
            i.column
            for i in self.add_columns
            if i.table_class_name == table_class_name
        ]

    @property
    def table_class_names(self) -> t.List[str]:
        return list({i.table_class_name for i in self.add_columns})


@dataclass
class DropColumnCollection:
    drop_columns: t.List[DropColumn] = field(default_factory=list)

    def append(self, drop_column: DropColumn):
        self.drop_columns.append(drop_column)

    def for_table_class_name(
        self, table_class_name: str
    ) -> t.List[DropColumn]:
        return [
            i
            for i in self.drop_columns
            if i.table_class_name == table_class_name
        ]

    @property
    def table_class_names(self) -> t.List[str]:
        return list({i.table_class_name for i in self.drop_columns})


@dataclass
class RenameColumnCollection:
    rename_columns: t.List[RenameColumn] = field(default_factory=list)

    def append(self, rename_column: RenameColumn):
        self.rename_columns.append(rename_column)

    def for_table_class_name(
        self, table_class_name: str
    ) -> t.List[RenameColumn]:
        return [
            i
            for i in self.rename_columns
            if i.table_class_name == table_class_name
        ]

    @property
    def table_class_names(self) -> t.List[str]:
        return list({i.table_class_name for i in self.rename_columns})


@dataclass
class AlterColumnCollection:
    alter_columns: t.List[AlterColumn] = field(default_factory=list)

    def append(self, alter_column: AlterColumn):
        self.alter_columns.append(alter_column)

    def for_table_class_name(
        self, table_class_name: str
    ) -> t.List[AlterColumn]:
        return [
            i
            for i in self.alter_columns
            if i.table_class_name == table_class_name
        ]

    @property
    def table_class_names(self) -> t.List[str]:
        return list({i.table_class_name for i in self.alter_columns})


AsyncFunction = t.Callable[[], t.Coroutine]


@dataclass
class MigrationManager:
    """
    Each auto generated migration returns a MigrationManager. It contains
    all of the schema changes that migration wants to make.
    """

    migration_id: str = ""
    app_name: str = ""
    description: str = ""
    preview: bool = False
    add_tables: t.List[DiffableTable] = field(default_factory=list)
    drop_tables: t.List[DiffableTable] = field(default_factory=list)
    rename_tables: t.List[RenameTable] = field(default_factory=list)
    change_table_schemas: t.List[ChangeTableSchema] = field(
        default_factory=list
    )
    add_columns: AddColumnCollection = field(
        default_factory=AddColumnCollection
    )
    drop_columns: DropColumnCollection = field(
        default_factory=DropColumnCollection
    )
    rename_columns: RenameColumnCollection = field(
        default_factory=RenameColumnCollection
    )
    alter_columns: AlterColumnCollection = field(
        default_factory=AlterColumnCollection
    )
    raw: t.List[t.Union[t.Callable, AsyncFunction]] = field(
        default_factory=list
    )
    raw_backwards: t.List[t.Union[t.Callable, AsyncFunction]] = field(
        default_factory=list
    )
    fake: bool = False

    def add_table(
        self,
        class_name: str,
        tablename: str,
        schema: t.Optional[str] = None,
        columns: t.Optional[t.List[Column]] = None,
    ):
        if not columns:
            columns = []

        self.add_tables.append(
            DiffableTable(
                class_name=class_name,
                tablename=tablename,
                columns=columns,
                schema=schema,
            )
        )

    def drop_table(
        self, class_name: str, tablename: str, schema: t.Optional[str] = None
    ):
        self.drop_tables.append(
            DiffableTable(
                class_name=class_name, tablename=tablename, schema=schema
            )
        )

    def change_table_schema(
        self,
        class_name: str,
        tablename: str,
        new_schema: t.Optional[str] = None,
        old_schema: t.Optional[str] = None,
    ):
        self.change_table_schemas.append(
            ChangeTableSchema(
                class_name=class_name,
                tablename=tablename,
                new_schema=new_schema,
                old_schema=old_schema,
            )
        )

    def rename_table(
        self,
        old_class_name: str,
        old_tablename: str,
        new_class_name: str,
        new_tablename: str,
        schema: t.Optional[str] = None,
    ):
        self.rename_tables.append(
            RenameTable(
                old_class_name=old_class_name,
                old_tablename=old_tablename,
                new_class_name=new_class_name,
                new_tablename=new_tablename,
                schema=schema,
            )
        )

    def add_column(
        self,
        table_class_name: str,
        tablename: str,
        column_name: str,
        db_column_name: t.Optional[str] = None,
        column_class_name: str = "",
        column_class: t.Optional[t.Type[Column]] = None,
        params: t.Optional[t.Dict[str, t.Any]] = None,
        schema: t.Optional[str] = None,
    ):
        """
        Add a new column to the table.

        :param column_class_name:
            The column type was traditionally specified as a string, using this
            variable. This didn't allow users to define custom column types
            though, which is why newer migrations directly reference a
            ``Column`` subclass using ``column_class``.
        :param column_class:
            A direct reference to a ``Column`` subclass.

        """
        if params is None:
            params = {}
        column_class = column_class or getattr(column_types, column_class_name)

        if column_class is None:
            raise ValueError("Unrecognised column type")

        cleaned_params = deserialise_params(params=params)
        column = column_class(**cleaned_params)
        column._meta.name = column_name
        if db_column_name:
            column._meta.db_column_name = db_column_name

        self.add_columns.append(
            AddColumnClass(
                column=column,
                tablename=tablename,
                table_class_name=table_class_name,
                schema=schema,
            )
        )

    def drop_column(
        self,
        table_class_name: str,
        tablename: str,
        column_name: str,
        db_column_name: t.Optional[str] = None,
        schema: t.Optional[str] = None,
    ):
        self.drop_columns.append(
            DropColumn(
                table_class_name=table_class_name,
                column_name=column_name,
                db_column_name=db_column_name or column_name,
                tablename=tablename,
                schema=schema,
            )
        )

    def rename_column(
        self,
        table_class_name: str,
        tablename: str,
        old_column_name: str,
        new_column_name: str,
        old_db_column_name: t.Optional[str] = None,
        new_db_column_name: t.Optional[str] = None,
        schema: t.Optional[str] = None,
    ):
        self.rename_columns.append(
            RenameColumn(
                table_class_name=table_class_name,
                tablename=tablename,
                old_column_name=old_column_name,
                new_column_name=new_column_name,
                old_db_column_name=old_db_column_name or old_column_name,
                new_db_column_name=new_db_column_name or new_column_name,
                schema=schema,
            )
        )

    def alter_column(
        self,
        table_class_name: str,
        tablename: str,
        column_name: str,
        db_column_name: t.Optional[str] = None,
        params: t.Optional[t.Dict[str, t.Any]] = None,
        old_params: t.Optional[t.Dict[str, t.Any]] = None,
        column_class: t.Optional[t.Type[Column]] = None,
        old_column_class: t.Optional[t.Type[Column]] = None,
        schema: t.Optional[str] = None,
    ):
        """
        All possible alterations aren't currently supported.
        """
        if params is None:
            params = {}
        if old_params is None:
            old_params = {}
        self.alter_columns.append(
            AlterColumn(
                table_class_name=table_class_name,
                tablename=tablename,
                column_name=column_name,
                db_column_name=db_column_name or column_name,
                params=params,
                old_params=old_params,
                column_class=column_class,
                old_column_class=old_column_class,
                schema=schema,
            )
        )

    def add_raw(self, raw: t.Union[t.Callable, AsyncFunction]):
        """
        A migration manager can execute arbitrary functions or coroutines when
        run. This is useful if you want to execute raw SQL.
        """
        self.raw.append(raw)

    def add_raw_backwards(self, raw: t.Union[t.Callable, AsyncFunction]):
        """
        When reversing a migration, you may want to run extra code to help
        clean up.
        """
        self.raw_backwards.append(raw)

    ###########################################################################

    async def get_table_from_snapshot(
        self,
        table_class_name: str,
        app_name: t.Optional[str],
        offset: int = 0,
        migration_id: t.Optional[str] = None,
    ) -> t.Type[Table]:
        """
        Returns a Table subclass which can be used for modifying data within
        a migration.

        :param offset:
            Lets you get a table as it appeared in an older migration. If the
            offset is -1, the table will come from the previous migration.

        """
        from piccolo.apps.migrations.commands.base import BaseMigrationManager

        if migration_id is None:
            migration_id = self.migration_id

        if app_name is None:
            app_name = self.app_name

        diffable_table = await BaseMigrationManager().get_table_from_snapshot(
            app_name=app_name,
            table_class_name=table_class_name,
            max_migration_id=migration_id,
            offset=offset,
        )
        return diffable_table.to_table_class()

    ###########################################################################

    @staticmethod
    async def _print_query(query: t.Union[DDL, Query, SchemaDDLBase]):
        if isinstance(query, DDL):
            print("\n", ";".join(query.ddl) + ";")
        else:
            print(str(query))

    async def _run_query(self, query: t.Union[DDL, Query, SchemaDDLBase]):
        """
        If MigrationManager is not in the preview mode,
         executes the queries. else, prints the query.
        """
        if self.preview:
            await self._print_query(query)
        else:
            await query.run()

    async def _run_alter_columns(self, backwards: bool = False):
        for table_class_name in self.alter_columns.table_class_names:
            alter_columns = self.alter_columns.for_table_class_name(
                table_class_name
            )

            if not alter_columns:
                continue

            _Table: t.Type[Table] = create_table_class(
                class_name=table_class_name,
                class_kwargs={
                    "tablename": alter_columns[0].tablename,
                    "schema": alter_columns[0].schema,
                },
            )

            for alter_column in alter_columns:
                params = (
                    alter_column.old_params
                    if backwards
                    else alter_column.params
                )

                old_params = (
                    alter_column.params
                    if backwards
                    else alter_column.old_params
                )

                ###############################################################

                # Change the column type if possible
                column_class = (
                    alter_column.old_column_class
                    if backwards
                    else alter_column.column_class
                )
                old_column_class = (
                    alter_column.column_class
                    if backwards
                    else alter_column.old_column_class
                )

                if (old_column_class is not None) and (
                    column_class is not None
                ):
                    if old_column_class != column_class:
                        old_column = old_column_class(**old_params)
                        old_column._meta._table = _Table
                        old_column._meta._name = alter_column.column_name
                        old_column._meta.db_column_name = (
                            alter_column.db_column_name
                        )

                        new_column = column_class(**params)
                        new_column._meta._table = _Table
                        new_column._meta._name = alter_column.column_name
                        new_column._meta.db_column_name = (
                            alter_column.db_column_name
                        )

                        using_expression: t.Optional[str] = None

                        # Postgres won't automatically cast some types to
                        # others. We may as well try, as it will definitely
                        # fail otherwise.
                        if new_column.value_type != old_column.value_type:
                            if old_params.get("default", ...) is not None:
                                # Unless the column's default value is also
                                # something which can be cast to the new type,
                                # it will also fail. Drop the default value for
                                # now - the proper default is set later on.
                                await self._run_query(
                                    _Table.alter().drop_default(old_column)
                                )

                            using_expression = "{}::{}".format(
                                alter_column.db_column_name,
                                new_column.column_type,
                            )

                        # We can't migrate a SERIAL to a BIGSERIAL or vice
                        # versa, as SERIAL isn't a true type, just an alias to
                        # other commands.
                        if issubclass(column_class, Serial) and issubclass(
                            old_column_class, Serial
                        ):
                            colored_warning(
                                "Unable to migrate Serial to BigSerial and "
                                "vice versa. This must be done manually."
                            )
                        else:
                            await self._run_query(
                                _Table.alter().set_column_type(
                                    old_column=old_column,
                                    new_column=new_column,
                                    using_expression=using_expression,
                                )
                            )

                ###############################################################

                null = params.get("null")
                if null is not None:
                    await self._run_query(
                        _Table.alter().set_null(
                            column=alter_column.db_column_name, boolean=null
                        )
                    )

                length = params.get("length")
                if length is not None:
                    await self._run_query(
                        _Table.alter().set_length(
                            column=alter_column.db_column_name, length=length
                        )
                    )

                unique = params.get("unique")
                if unique is not None:
                    # When modifying unique constraints, we need to pass in
                    # a column type, and not just the column name.
                    column = Column()
                    column._meta._table = _Table
                    column._meta._name = alter_column.column_name
                    column._meta.db_column_name = alter_column.db_column_name
                    await self._run_query(
                        _Table.alter().set_unique(
                            column=column, boolean=unique
                        )
                    )

                index = params.get("index")
                index_method = params.get("index_method")
                if index is None:
                    if index_method is not None:
                        # If the index value hasn't changed, but the
                        # index_method value has, this indicates we need
                        # to change the index type.
                        column = Column()
                        column._meta._table = _Table
                        column._meta._name = alter_column.column_name
                        column._meta.db_column_name = (
                            alter_column.db_column_name
                        )
                        await self._run_query(_Table.drop_index([column]))
                        await self._run_query(
                            _Table.create_index(
                                [column],
                                method=index_method,
                                if_not_exists=True,
                            )
                        )
                else:
                    # If the index value has changed, then we are either
                    # dropping, or creating an index.
                    column = Column()
                    column._meta._table = _Table
                    column._meta._name = alter_column.column_name
                    column._meta.db_column_name = alter_column.db_column_name

                    if index is True:
                        kwargs = (
                            {"method": index_method} if index_method else {}
                        )
                        await self._run_query(
                            _Table.create_index(
                                [column], if_not_exists=True, **kwargs
                            )
                        )
                    else:
                        await self._run_query(_Table.drop_index([column]))

                # None is a valid value, so retrieve ellipsis if not found.
                default = params.get("default", ...)
                if default is not ...:
                    column = Column()
                    column._meta._table = _Table
                    column._meta._name = alter_column.column_name
                    column._meta.db_column_name = alter_column.db_column_name

                    if default is None:
                        await self._run_query(
                            _Table.alter().drop_default(column=column)
                        )
                    else:
                        column.default = default
                        await self._run_query(
                            _Table.alter().set_default(
                                column=column, value=column.get_default_value()
                            )
                        )

                # None is a valid value, so retrieve ellipsis if not found.
                digits = params.get("digits", ...)
                if digits is not ...:
                    await self._run_query(
                        _Table.alter().set_digits(
                            column=alter_column.db_column_name,
                            digits=digits,
                        )
                    )

    async def _run_drop_tables(self, backwards=False):
        for diffable_table in self.drop_tables:
            if backwards:
                _Table = await self.get_table_from_snapshot(
                    table_class_name=diffable_table.class_name,
                    app_name=self.app_name,
                    offset=-1,
                )
                await self._run_query(_Table.create_table())
            else:
                await self._run_query(
                    diffable_table.to_table_class().alter().drop_table()
                )

    async def _run_drop_columns(self, backwards: bool = False):
        if backwards:
            for drop_column in self.drop_columns.drop_columns:
                _Table = await self.get_table_from_snapshot(
                    table_class_name=drop_column.table_class_name,
                    app_name=self.app_name,
                    offset=-1,
                )
                column_to_restore = _Table._meta.get_column_by_name(
                    drop_column.column_name
                )
                await self._run_query(
                    _Table.alter().add_column(
                        name=drop_column.column_name, column=column_to_restore
                    )
                )
        else:
            for table_class_name in self.drop_columns.table_class_names:
                columns = self.drop_columns.for_table_class_name(
                    table_class_name
                )

                if not columns:
                    continue

                _Table = create_table_class(
                    class_name=table_class_name,
                    class_kwargs={
                        "tablename": columns[0].tablename,
                        "schema": columns[0].schema,
                    },
                )

                for column in columns:
                    await self._run_query(
                        _Table.alter().drop_column(column=column.column_name)
                    )

    async def _run_rename_tables(self, backwards: bool = False):
        for rename_table in self.rename_tables:
            class_name = (
                rename_table.new_class_name
                if backwards
                else rename_table.old_class_name
            )
            tablename = (
                rename_table.new_tablename
                if backwards
                else rename_table.old_tablename
            )
            new_tablename = (
                rename_table.old_tablename
                if backwards
                else rename_table.new_tablename
            )

            _Table: t.Type[Table] = create_table_class(
                class_name=class_name,
                class_kwargs={
                    "tablename": tablename,
                    "schema": rename_table.schema,
                },
            )

            await self._run_query(
                _Table.alter().rename_table(new_name=new_tablename)
            )

    async def _run_rename_columns(self, backwards: bool = False):
        for table_class_name in self.rename_columns.table_class_names:
            columns = self.rename_columns.for_table_class_name(
                table_class_name
            )

            if not columns:
                continue

            _Table: t.Type[Table] = create_table_class(
                class_name=table_class_name,
                class_kwargs={
                    "tablename": columns[0].tablename,
                    "schema": columns[0].schema,
                },
            )

            for rename_column in columns:
                column = (
                    rename_column.new_db_column_name
                    if backwards
                    else rename_column.old_db_column_name
                )
                new_name = (
                    rename_column.old_db_column_name
                    if backwards
                    else rename_column.new_db_column_name
                )

                await self._run_query(
                    _Table.alter().rename_column(
                        column=column,
                        new_name=new_name,
                    )
                )

    async def _run_add_tables(self, backwards: bool = False):
        table_classes: t.List[t.Type[Table]] = []
        for add_table in self.add_tables:
            add_columns: t.List[AddColumnClass] = (
                self.add_columns.for_table_class_name(add_table.class_name)
            )
            _Table: t.Type[Table] = create_table_class(
                class_name=add_table.class_name,
                class_kwargs={
                    "tablename": add_table.tablename,
                    "schema": add_table.schema,
                },
                class_members={
                    add_column.column._meta.name: add_column.column
                    for add_column in add_columns
                },
            )
            table_classes.append(_Table)

        # Sort by foreign key, so they're created in the right order.
        sorted_table_classes = sort_table_classes(table_classes)

        if backwards:
            for _Table in reversed(sorted_table_classes):
                await self._run_query(_Table.alter().drop_table(cascade=True))
        else:
            for _Table in sorted_table_classes:
                await self._run_query(_Table.create_table())

    async def _run_add_columns(self, backwards: bool = False):
        """
        Add columns, which belong to existing tables
        """
        if backwards:
            for add_column in self.add_columns.add_columns:
                if add_column.table_class_name in [
                    i.class_name for i in self.add_tables
                ]:
                    # Don't reverse the add column as the table is going to
                    # be deleted.
                    continue

                _Table = create_table_class(
                    class_name=add_column.table_class_name,
                    class_kwargs={
                        "tablename": add_column.tablename,
                        "schema": add_column.schema,
                    },
                )

                await self._run_query(
                    _Table.alter().drop_column(add_column.column)
                )
        else:
            for table_class_name in self.add_columns.table_class_names:
                if table_class_name in [i.class_name for i in self.add_tables]:
                    continue  # No need to add columns to new tables

                add_columns: t.List[AddColumnClass] = (
                    self.add_columns.for_table_class_name(table_class_name)
                )

                ###############################################################
                # Define the table, with the columns, so the metaclass
                # sets up the columns correctly.

                table_class_members = {
                    add_column.column._meta.name: add_column.column
                    for add_column in add_columns
                }

                # There's an extreme edge case, when we're adding a foreign
                # key which references its own table, for example:
                #
                #   fk = ForeignKey('self')
                #
                # And that table has a custom primary key, for example:
                #
                #   id = UUID(primary_key=True)
                #
                # In this situation, we need to know the primary key of the
                # table in order to correctly add this new foreign key.
                for add_column in add_columns:
                    if (
                        isinstance(add_column.column, ForeignKey)
                        and add_column.column._meta.params.get("references")
                        == "self"
                    ):
                        try:
                            existing_table = (
                                await self.get_table_from_snapshot(
                                    table_class_name=table_class_name,
                                    app_name=self.app_name,
                                    offset=-1,
                                )
                            )
                        except ValueError:
                            logger.error(
                                "Unable to find primary key for the table - "
                                "assuming Serial."
                            )
                        else:
                            primary_key = existing_table._meta.primary_key

                            table_class_members[primary_key._meta.name] = (
                                primary_key
                            )

                        break

                _Table = create_table_class(
                    class_name=add_columns[0].table_class_name,
                    class_kwargs={
                        "tablename": add_columns[0].tablename,
                        "schema": add_columns[0].schema,
                    },
                    class_members=table_class_members,
                )

                ###############################################################

                for add_column in add_columns:
                    # We fetch the column from the Table, as the metaclass
                    # copies and sets it up properly.
                    column = _Table._meta.get_column_by_name(
                        add_column.column._meta.name
                    )

                    await self._run_query(
                        _Table.alter().add_column(
                            name=column._meta.name, column=column
                        )
                    )
                    if add_column.column._meta.index:
                        await self._run_query(
                            _Table.create_index([add_column.column])
                        )

    async def _run_change_table_schema(self, backwards: bool = False):
        from piccolo.schema import SchemaManager

        schema_manager = SchemaManager()

        for change_table_schema in self.change_table_schemas:
            if backwards:
                # Note, we don't try dropping any schemas we may have created.
                # It's dangerous to do so, just in case the user manually
                # added tables etc to the schema, and we delete them.

                if (
                    change_table_schema.old_schema
                    and change_table_schema.old_schema != "public"
                ):
                    await self._run_query(
                        schema_manager.create_schema(
                            schema_name=change_table_schema.old_schema,
                            if_not_exists=True,
                        )
                    )

                await self._run_query(
                    schema_manager.move_table(
                        table_name=change_table_schema.tablename,
                        new_schema=change_table_schema.old_schema or "public",
                        current_schema=change_table_schema.new_schema,
                    )
                )

            else:
                if (
                    change_table_schema.new_schema
                    and change_table_schema.new_schema != "public"
                ):
                    await self._run_query(
                        schema_manager.create_schema(
                            schema_name=change_table_schema.new_schema,
                            if_not_exists=True,
                        )
                    )

                await self._run_query(
                    schema_manager.move_table(
                        table_name=change_table_schema.tablename,
                        new_schema=change_table_schema.new_schema or "public",
                        current_schema=change_table_schema.old_schema,
                    )
                )

    async def run(self, backwards: bool = False):
        direction = "backwards" if backwards else "forwards"
        if self.preview:
            direction = "preview " + direction
        print(f"  - {self.migration_id} [{direction}]... ", end="")

        engine = engine_finder()

        if not engine:
            raise Exception("Can't find engine")

        async with engine.transaction():
            if not self.preview:
                if direction == "backwards":
                    raw_list = self.raw_backwards
                else:
                    raw_list = self.raw

                for raw in raw_list:
                    if inspect.iscoroutinefunction(raw):
                        await raw()
                    else:
                        raw()

            await self._run_add_tables(backwards=backwards)
            await self._run_change_table_schema(backwards=backwards)
            await self._run_rename_tables(backwards=backwards)
            await self._run_add_columns(backwards=backwards)
            await self._run_drop_columns(backwards=backwards)
            await self._run_drop_tables(backwards=backwards)
            await self._run_rename_columns(backwards=backwards)
            # We can remove this for cockroach when resolved.
            # https://github.com/cockroachdb/cockroach/issues/49351
            # "ALTER COLUMN TYPE is not supported inside a transaction"
            if engine.engine_type != "cockroach":
                await self._run_alter_columns(backwards=backwards)

        if engine.engine_type == "cockroach":
            await self._run_alter_columns(backwards=backwards)



import typing as t

from piccolo.columns.base import Column
from piccolo.querystring import QueryString


class Function(QueryString):
    function_name: str

    def __init__(
        self,
        identifier: t.Union[Column, QueryString, str],
        alias: t.Optional[str] = None,
    ):
        alias = alias or self.__class__.__name__.lower()

        super().__init__(
            f"{self.function_name}({{}})",
            identifier,
            alias=alias,
        )



from __future__ import annotations

import sys
import typing as t
from unittest import TestCase
from unittest.mock import MagicMock, call, patch

from piccolo.apps.migrations.commands.backwards import backwards
from piccolo.apps.migrations.commands.forwards import forwards
from piccolo.apps.migrations.tables import Migration
from piccolo.utils.sync import run_sync
from tests.base import engines_only
from tests.example_apps.music.tables import (
    Band,
    Concert,
    Instrument,
    Manager,
    Poster,
    RecordingStudio,
    Shirt,
    Ticket,
    Venue,
)

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.table import Table

TABLE_CLASSES: t.List[t.Type[Table]] = [
    Manager,
    Band,
    Venue,
    Concert,
    Ticket,
    Poster,
    Shirt,
    RecordingStudio,
    Instrument,
]


@engines_only("postgres", "cockroach")
class TestForwardsBackwards(TestCase):
    """
    Test the forwards and backwards migration commands.
    """

    def test_forwards_backwards_all_migrations(self):
        """
        Test running all of the migrations forwards, then backwards.
        """
        for app_name in ("music", "all"):
            run_sync(forwards(app_name=app_name, migration_id="all"))

            # Check the tables exist
            for table_class in TABLE_CLASSES:
                self.assertTrue(table_class.table_exists().run_sync())
            self.assertNotEqual(Migration.count().run_sync(), 0)

            run_sync(
                backwards(
                    app_name=app_name, migration_id="all", auto_agree=True
                )
            )

            # Check the tables don't exist
            for table_class in TABLE_CLASSES:
                self.assertTrue(not table_class.table_exists().run_sync())
            self.assertEqual(Migration.count().run_sync(), 0)
            # Preview
            run_sync(
                forwards(app_name=app_name, migration_id="all", preview=True)
            )
            for table_class in TABLE_CLASSES:
                self.assertTrue(not table_class.table_exists().run_sync())
            self.assertEqual(Migration.count().run_sync(), 0)

    def test_forwards_backwards_single_migration(self):
        """
        Test running a single migrations forwards, then backwards.
        """
        table_classes = [Band, Manager]

        for migration_id in ["1", "2020-12-17T18:44:30"]:
            run_sync(forwards(app_name="music", migration_id=migration_id))

            # Check the tables exist
            for table_class in table_classes:
                self.assertTrue(table_class.table_exists().run_sync())

            self.assertTrue(
                Migration.exists()
                .where(Migration.name == "2020-12-17T18:44:30")
                .run_sync()
            )

            run_sync(
                backwards(
                    app_name="music",
                    migration_id=migration_id,
                    auto_agree=True,
                )
            )

            # Check the tables don't exist
            for table_class in table_classes:
                self.assertTrue(not table_class.table_exists().run_sync())
            self.assertFalse(
                Migration.exists()
                .where(Migration.name == "2020-12-17T18:44:30")
                .run_sync()
            )

            # Preview
            run_sync(
                forwards(
                    app_name="music", migration_id=migration_id, preview=True
                )
            )
            for table_class in table_classes:
                self.assertTrue(not table_class.table_exists().run_sync())
            self.assertFalse(
                Migration.exists()
                .where(Migration.name == "2020-12-17T18:44:30")
                .run_sync()
            )

    @patch("piccolo.apps.migrations.commands.forwards.print")
    def test_forwards_unknown_migration(self, print_: MagicMock):
        """
        Test running an unknown migrations forwards.
        """
        with self.assertRaises(SystemExit):
            run_sync(
                forwards(app_name="music", migration_id="migration-12345")
            )

        self.assertTrue(
            call("migration-12345 is unrecognised", file=sys.stderr)
            in print_.mock_calls
        )

    @patch("piccolo.apps.migrations.commands.backwards.print")
    def test_backwards_unknown_migration(self, print_: MagicMock):
        """
        Test running an unknown migrations backwards.
        """
        run_sync(forwards(app_name="music", migration_id="all"))

        with self.assertRaises(SystemExit):
            run_sync(
                backwards(
                    app_name="music",
                    migration_id="migration-12345",
                    auto_agree=True,
                )
            )

        self.assertTrue(
            tuple(print_.mock_calls[0])[1][0].startswith(
                "Unrecognized migration name - must be one of "
            )
        )

    @patch("piccolo.apps.migrations.commands.backwards.print")
    def test_backwards_no_migrations(self, print_: MagicMock):
        """
        Test running migrations backwards if none have been run previously.
        """
        run_sync(
            backwards(
                app_name="music",
                migration_id="2020-12-17T18:44:30",
                auto_agree=True,
            )
        )
        self.assertTrue(
            call("🏁 No migrations to reverse!") in print_.mock_calls
        )

    @patch("piccolo.apps.migrations.commands.forwards.print")
    def test_forwards_no_migrations(self, print_: MagicMock):
        """
        Test running the migrations if they've already run.
        """
        run_sync(forwards(app_name="music", migration_id="all"))
        run_sync(forwards(app_name="music", migration_id="all"))

        self.assertTrue(
            print_.mock_calls[-1] == call("🏁 No migrations need to be run")
        )

    @engines_only("postgres")
    def test_forwards_fake(self):
        """
        Make sure migrations can be faked on the command line.
        """
        run_sync(forwards(app_name="music", migration_id="all", fake=True))

        for table_class in TABLE_CLASSES:
            self.assertTrue(not table_class.table_exists().run_sync())

        ran_migration_names = (
            Migration.select(Migration.name).output(as_list=True).run_sync()
        )

        self.assertEqual(
            ran_migration_names,
            # TODO - rather than hardcoding, might fetch these dynamically.
            [
                "2020-12-17T18:44:30",
                "2020-12-17T18:44:39",
                "2020-12-17T18:44:44",
                "2021-07-25T22:38:48:009306",
                "2021-09-06T13:58:23:024723",
                "2021-11-13T14:01:46:114725",
                "2024-05-28T23:15:41:018844",
                "2024-06-19T18:11:05:793132",
            ],
        )

    @engines_only("postgres")
    @patch("piccolo.apps.migrations.commands.forwards.print")
    def test_hardcoded_fake_migrations(self, print_: MagicMock):
        """
        Make sure that migrations that have been hardcoded as fake aren't
        executed, even without the ``--fake`` command line flag.

        See tests/example_apps/music/piccolo_migrations/music_2024_06_19t18_11_05_793132.py

        """  # noqa: E501
        run_sync(forwards(app_name="music", migration_id="all"))

        # The migration which is hardcoded as fake:
        migration_name = "2024-06-19T18:11:05:793132"

        self.assertTrue(
            Migration.exists()
            .where(Migration.name == migration_name)
            .run_sync()
        )

        self.assertNotIn(
            call("Running fake migration"),
            print_.mock_calls,
        )
        self.assertIn(
            call(f"- {migration_name}: faked! ⏭️"),
            print_.mock_calls,
        )

    def tearDown(self):
        for table_class in TABLE_CLASSES + [Migration]:
            table_class.alter().drop_table(
                cascade=True, if_exists=True
            ).run_sync()



from __future__ import annotations

import datetime
import typing as t
from enum import Enum

from .base import Default


class TimeOffset(Default):
    def __init__(self, hours: int, minutes: int, seconds: int):
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds

    @property
    def postgres(self):
        interval_string = self.get_postgres_interval_string(
            ["hours", "minutes", "seconds"]
        )
        return f"CURRENT_TIME + INTERVAL '{interval_string}'"

    @property
    def cockroach(self):
        interval_string = self.get_postgres_interval_string(
            ["hours", "minutes", "seconds"]
        )
        return f"CURRENT_TIME::TIMESTAMP + INTERVAL '{interval_string}'"

    @property
    def sqlite(self):
        interval_string = self.get_sqlite_interval_string(
            ["hours", "minutes", "seconds"]
        )
        return f"(time(CURRENT_TIME, {interval_string}))"

    def python(self):
        return (
            datetime.datetime.now()
            + datetime.timedelta(
                hours=self.hours, minutes=self.minutes, seconds=self.seconds
            )
        ).time


class TimeNow(Default):
    @property
    def postgres(self):
        return "CURRENT_TIME"

    @property
    def cockroach(self):
        return "CURRENT_TIME::TIMESTAMP"

    @property
    def sqlite(self):
        return "CURRENT_TIME"

    def python(self):
        return datetime.datetime.now().time()


class TimeCustom(Default):
    def __init__(self, hour: int, minute: int, second: int):
        self.hour = hour
        self.minute = minute
        self.second = second
        self.time = datetime.time(hour=hour, minute=minute, second=second)

    @property
    def postgres(self):
        return f"'{self.time.isoformat()}'"

    @property
    def cockroach(self):
        return f"'{self.time.isoformat()}'::TIMESTAMP"

    @property
    def sqlite(self):
        return f"'{self.time.isoformat()}'"

    def python(self):
        return self.time

    @classmethod
    def from_time(cls, instance: datetime.time):
        return cls(
            hour=instance.hour, minute=instance.minute, second=instance.second
        )


TimeArg = t.Union[
    TimeCustom,
    TimeNow,
    TimeOffset,
    Enum,
    None,
    datetime.time,
    t.Callable[[], datetime.time],
]


__all__ = ["TimeArg", "TimeCustom", "TimeNow", "TimeOffset"]



from .base import Operator


class MathOperator(Operator):
    pass


class Add(MathOperator):
    template = "{name} + {value}"


class Subtract(MathOperator):
    template = "{name} - {value}"


class Multiply(MathOperator):
    template = "{name} * {value}"


class Divide(MathOperator):
    template = "{name} * {value}"



import sys
import typing as t
from getpass import getpass, getuser

from piccolo.apps.user.tables import BaseUser


def get_username() -> str:
    default_username = getuser()
    username = input(f"Enter username ({default_username}):\n")
    return default_username if not username else username


def get_email() -> str:
    return input("Enter email:\n")


def get_password() -> str:
    return getpass("Enter password:\n")


def get_confirmed_password() -> str:
    return getpass("Confirm password:\n")


def get_is_admin() -> bool:
    while True:
        admin = input("Admin user? Enter y or n:\n")
        if admin in ("y", "n"):
            break
        else:
            print("Unrecognised option")

    return admin == "y"


def get_is_superuser() -> bool:
    while True:
        superuser = input("Superuser? Enter y or n:\n")
        if superuser in ("y", "n"):
            break
        else:
            print("Unrecognised option")

    return superuser == "y"


def get_is_active() -> bool:
    while True:
        active = input("Active? Enter y or n:\n")
        if active in ("y", "n"):
            break
        else:
            print("Unrecognised option")

    return active == "y"


def create(
    username: t.Optional[str] = None,
    email: t.Optional[str] = None,
    password: t.Optional[str] = None,
    is_admin: t.Optional[bool] = None,
    is_superuser: t.Optional[bool] = None,
    is_active: t.Optional[bool] = None,
):
    """
    Create a new user.
    """
    username = get_username() if username is None else username
    email = get_email() if email is None else email
    if password is None:
        password = get_password()
        confirmed_password = get_confirmed_password()

        if password != confirmed_password:
            sys.exit("Passwords don't match!")

    is_admin = get_is_admin() if is_admin is None else is_admin
    is_superuser = get_is_superuser() if is_superuser is None else is_superuser
    is_active = get_is_active() if is_active is None else is_active

    user = BaseUser.create_user_sync(
        username=username,
        password=password,
        admin=is_admin,
        email=email,
        active=is_active,
        superuser=is_superuser,
    )

    print(f"Created User {user.id}")



from piccolo import columns
from piccolo.columns.readable import Readable
from piccolo.table import Table
from tests.base import TableTest


class MyTable(Table):
    first_name = columns.Varchar()
    last_name = columns.Varchar()

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(
            template="%s %s", columns=[cls.first_name, cls.last_name]
        )


class TestReadable(TableTest):
    tables = [MyTable]

    def setUp(self):
        super().setUp()
        MyTable(first_name="Guido", last_name="van Rossum").save().run_sync()

    def test_readable(self):
        response = MyTable.select(MyTable.get_readable()).run_sync()
        self.assertEqual(response[0]["readable"], "Guido van Rossum")



import decimal

from piccolo.columns import Numeric
from piccolo.query.functions.math import Abs, Ceil, Floor, Round
from piccolo.table import Table
from tests.base import TableTest


class Ticket(Table):
    price = Numeric(digits=(5, 2))


class TestMath(TableTest):

    tables = [Ticket]

    def setUp(self):
        super().setUp()
        self.ticket = Ticket({Ticket.price: decimal.Decimal("36.50")})
        self.ticket.save().run_sync()

    def test_floor(self):
        response = Ticket.select(Floor(Ticket.price, alias="price")).run_sync()
        self.assertListEqual(response, [{"price": decimal.Decimal("36.00")}])

    def test_ceil(self):
        response = Ticket.select(Ceil(Ticket.price, alias="price")).run_sync()
        self.assertListEqual(response, [{"price": decimal.Decimal("37.00")}])

    def test_abs(self):
        self.ticket.price = decimal.Decimal("-1.50")
        self.ticket.save().run_sync()
        response = Ticket.select(Abs(Ticket.price, alias="price")).run_sync()
        self.assertListEqual(response, [{"price": decimal.Decimal("1.50")}])

    def test_round(self):
        response = Ticket.select(Round(Ticket.price, alias="price")).run_sync()
        self.assertListEqual(response, [{"price": decimal.Decimal("37.00")}])



"""
These functions mirror their counterparts in the Postgresql docs:

https://www.postgresql.org/docs/current/functions-string.html

"""

import typing as t

from piccolo.columns.base import Column
from piccolo.columns.column_types import Text, Varchar
from piccolo.querystring import QueryString

from .base import Function


class Length(Function):
    """
    Returns the number of characters in the string.
    """

    function_name = "LENGTH"


class Lower(Function):
    """
    Converts the string to all lower case, according to the rules of the
    database's locale.
    """

    function_name = "LOWER"


class Ltrim(Function):
    """
    Removes the longest string containing only characters in characters (a
    space by default) from the start of string.
    """

    function_name = "LTRIM"


class Reverse(Function):
    """
    Return reversed string.

    Not supported in SQLite.

    """

    function_name = "REVERSE"


class Rtrim(Function):
    """
    Removes the longest string containing only characters in characters (a
    space by default) from the end of string.
    """

    function_name = "RTRIM"


class Upper(Function):
    """
    Converts the string to all upper case, according to the rules of the
    database's locale.
    """

    function_name = "UPPER"


class Concat(QueryString):
    def __init__(
        self,
        *args: t.Union[Column, QueryString, str],
        alias: t.Optional[str] = None,
    ):
        """
        Concatenate multiple values into a single string.

        .. note::
            Null values are ignored, so ``null + '!!!'`` returns ``!!!``,
            not ``null``.

        .. warning::
            For SQLite, this is only available in version 3.44.0 and above.

        """
        if len(args) < 2:
            raise ValueError("At least two values must be passed in.")

        placeholders = ", ".join("{}" for _ in args)

        processed_args: t.List[t.Union[QueryString, Column]] = []

        for arg in args:
            if isinstance(arg, str) or (
                isinstance(arg, Column)
                and not isinstance(arg, (Varchar, Text))
            ):
                processed_args.append(QueryString("CAST({} AS TEXT)", arg))
            else:
                processed_args.append(arg)

        super().__init__(
            f"CONCAT({placeholders})", *processed_args, alias=alias
        )


__all__ = (
    "Length",
    "Lower",
    "Ltrim",
    "Reverse",
    "Rtrim",
    "Upper",
    "Concat",
)



import typing as t

from piccolo.apps.user.tables import BaseUser
from piccolo.columns import Column
from piccolo.utils.printing import print_dict_table

ORDER_BY_COLUMN_NAMES = [
    i._meta.name for i in BaseUser.all_columns(exclude=[BaseUser.password])
]


async def get_users(
    order_by: Column, ascending: bool, limit: int, page: int
) -> t.List[t.Dict[str, t.Any]]:
    return (
        await BaseUser.select(
            *BaseUser.all_columns(exclude=[BaseUser.password])
        )
        .order_by(
            order_by,
            ascending=ascending,
        )
        .limit(limit)
        .offset(limit * (page - 1))
    )


async def list_users(
    limit: int = 20, page: int = 1, order_by: str = "username"
):
    """
    List existing users.

    :param limit:
        The maximum number of users to list.
    :param page:
        Lets you paginate through the list of users.
    :param order_by:
        The column used to order the results. Prefix with '-' for descending
        order.

    """
    if page < 1:
        raise ValueError("The page number must > 0.")

    if limit < 1:
        raise ValueError("The limit number must be > 0.")

    ascending = True
    if order_by.startswith("-"):
        ascending = False
        order_by = order_by[1:]

    if order_by not in ORDER_BY_COLUMN_NAMES:
        raise ValueError(
            "The order_by argument must be one of the following: "
            + ", ".join(ORDER_BY_COLUMN_NAMES)
        )

    users = await get_users(
        order_by=BaseUser._meta.get_column_by_name(order_by),
        ascending=ascending,
        limit=limit,
        page=page,
    )

    if len(users) == 0:
        print("No data")
        return

    print_dict_table(users, header_separator=True)



from unittest import TestCase

from piccolo.columns import Varchar
from piccolo.schema import SchemaManager
from piccolo.table import Table
from tests.base import engines_only
from tests.example_apps.music.tables import Manager


class TestCreate(TestCase):
    def test_create_table(self):
        Manager.create_table().run_sync()
        self.assertTrue(Manager.table_exists().run_sync())

        Manager.alter().drop_table().run_sync()


class BandMember(Table):
    name = Varchar(length=50, index=True)


class TestCreateWithIndexes(TestCase):
    def setUp(self):
        BandMember.create_table().run_sync()

    def tearDown(self):
        BandMember.alter().drop_table().run_sync()

    def test_create_table_with_indexes(self):
        index_names = BandMember.indexes().run_sync()
        index_name = BandMember._get_index_name(["name"])
        self.assertIn(index_name, index_names)

    def test_create_if_not_exists_with_indexes(self):
        """
        Make sure that if the same table is created again, with the
        `if_not_exists` flag, then no errors are raised for duplicate indexes
        (i.e. the indexes should also be created with IF NOT EXISTS).
        """
        query = BandMember.create_table(if_not_exists=True)

        # Shouldn't raise any errors:
        query.run_sync()

        self.assertTrue(
            query.ddl[0].__str__().startswith("CREATE TABLE IF NOT EXISTS"),
            query.ddl[1].__str__().startswith("CREATE INDEX IF NOT EXISTS"),
        )


@engines_only("postgres", "cockroach")
class TestCreateWithSchema(TestCase):

    manager = SchemaManager()

    class Band(Table, tablename="band", schema="schema_1"):
        name = Varchar(length=50, index=True)

    def tearDown(self) -> None:
        self.manager.drop_schema(
            schema_name="schema_1", cascade=True
        ).run_sync()

    def test_table_created(self):
        """
        Make sure that tables can be created in specific schemas.
        """
        Band = self.Band
        Band.create_table().run_sync()

        self.assertIn(
            "band",
            self.manager.list_tables(schema_name="schema_1").run_sync(),
        )


@engines_only("postgres", "cockroach")
class TestCreateWithPublicSchema(TestCase):
    class Band(Table, tablename="band", schema="public"):
        name = Varchar(length=50, index=True)

    def tearDown(self) -> None:
        self.Band.alter().drop_table(if_exists=True).run_sync()

    def test_table_created(self):
        """
        Make sure that if the schema is explicitly set a 'public' rather than
        ``None``, that we don't try creating the schema which would cause it
        to fail.
        """
        Band = self.Band
        Band.create_table().run_sync()

        self.assertIn(
            "band",
            SchemaManager().list_tables(schema_name="public").run_sync(),
        )



from .aggregate import Avg, Count, Max, Min, Sum
from .datetime import Day, Extract, Hour, Month, Second, Strftime, Year
from .math import Abs, Ceil, Floor, Round
from .string import Concat, Length, Lower, Ltrim, Reverse, Rtrim, Upper
from .type_conversion import Cast

__all__ = (
    "Abs",
    "Avg",
    "Cast",
    "Ceil",
    "Concat",
    "Count",
    "Day",
    "Extract",
    "Extract",
    "Floor",
    "Hour",
    "Length",
    "Lower",
    "Ltrim",
    "Max",
    "Min",
    "Month",
    "Reverse",
    "Round",
    "Rtrim",
    "Second",
    "Strftime",
    "Sum",
    "Upper",
    "Year",
)



from __future__ import annotations

import datetime
import json
import typing as t
from decimal import Decimal
from uuid import UUID

from piccolo.columns import JSON, JSONB, Array, Column, ForeignKey
from piccolo.custom_types import TableInstance
from piccolo.testing.random_builder import RandomBuilder
from piccolo.utils.sync import run_sync


class ModelBuilder:
    __DEFAULT_MAPPER: t.Dict[t.Type, t.Callable] = {
        bool: RandomBuilder.next_bool,
        bytes: RandomBuilder.next_bytes,
        datetime.date: RandomBuilder.next_date,
        datetime.datetime: RandomBuilder.next_datetime,
        float: RandomBuilder.next_float,
        int: RandomBuilder.next_int,
        str: RandomBuilder.next_str,
        datetime.time: RandomBuilder.next_time,
        datetime.timedelta: RandomBuilder.next_timedelta,
        UUID: RandomBuilder.next_uuid,
    }

    @classmethod
    async def build(
        cls,
        table_class: t.Type[TableInstance],
        defaults: t.Optional[t.Dict[t.Union[Column, str], t.Any]] = None,
        persist: bool = True,
        minimal: bool = False,
    ) -> TableInstance:
        """
        Build a ``Table`` instance with random data and save async.
        If the ``Table`` has any foreign keys, then the related rows are also
        created automatically.

        :param table_class:
            Table class to randomize.
        :param defaults:
            Any values specified here will be used instead of random values.
        :param persist:
            Whether to save the new instance in the database.
        :param minimal:
            If ``True`` then any columns with ``null=True`` are assigned
            a value of ``None``.

        Examples::

            # Create a new instance with all random values:
            manager = await ModelBuilder.build(Manager)

            # Create a new instance, with certain defaults:
            manager = await ModelBuilder.build(
                Manager,
                {Manager.name: 'Guido'}
            )

            # Create a new instance, but don't save it in the database:
            manager = await ModelBuilder.build(Manager, persist=False)

            # Create a new instance, with all null values set to None:
            manager = await ModelBuilder.build(Manager, minimal=True)

            # We can pass other table instances in as default values:
            band = await ModelBuilder.build(Band, {Band.manager: manager})

        """
        return await cls._build(
            table_class=table_class,
            defaults=defaults,
            persist=persist,
            minimal=minimal,
        )

    @classmethod
    def build_sync(
        cls,
        table_class: t.Type[TableInstance],
        defaults: t.Optional[t.Dict[t.Union[Column, str], t.Any]] = None,
        persist: bool = True,
        minimal: bool = False,
    ) -> TableInstance:
        """
        A sync wrapper around :meth:`build`.
        """
        return run_sync(
            cls.build(
                table_class=table_class,
                defaults=defaults,
                persist=persist,
                minimal=minimal,
            )
        )

    @classmethod
    async def _build(
        cls,
        table_class: t.Type[TableInstance],
        defaults: t.Optional[t.Dict[t.Union[Column, str], t.Any]] = None,
        minimal: bool = False,
        persist: bool = True,
    ) -> TableInstance:
        model = table_class(_ignore_missing=True)
        defaults = {} if not defaults else defaults

        for column, value in defaults.items():
            if isinstance(column, str):
                column = model._meta.get_column_by_name(column)

            setattr(model, column._meta.name, value)

        for column in model._meta.columns:
            if column._meta.null and minimal:
                continue

            if column._meta.name in defaults:
                continue  # Column value exists

            if isinstance(column, ForeignKey) and persist:
                # Check for recursion
                if column._foreign_key_meta.references is table_class:
                    if column._meta.null is True:
                        # We can avoid this problem entirely by setting it to
                        # None.
                        random_value = None
                    else:
                        # There's no way to avoid recursion in the situation.
                        raise ValueError("Recursive foreign key detected")
                else:
                    reference_model = await cls._build(
                        column._foreign_key_meta.resolved_references,
                        persist=True,
                    )
                    random_value = getattr(
                        reference_model,
                        reference_model._meta.primary_key._meta.name,
                    )
            else:
                random_value = cls._randomize_attribute(column)

            setattr(model, column._meta.name, random_value)

        if persist:
            await model.save().run()

        return model

    @classmethod
    def _randomize_attribute(cls, column: Column) -> t.Any:
        """
        Generate a random value for a column and apply formatting.

        :param column:
            Column class to randomize.

        """
        random_value: t.Any
        if column.value_type == Decimal:
            precision, scale = column._meta.params["digits"] or (4, 2)
            random_value = RandomBuilder.next_float(
                maximum=10 ** (precision - scale), scale=scale
            )
        elif column.value_type == datetime.datetime:
            tz_aware = getattr(column, "tz_aware", False)
            random_value = RandomBuilder.next_datetime(tz_aware=tz_aware)
        elif column.value_type == list:
            length = RandomBuilder.next_int(maximum=10)
            base_type = t.cast(Array, column).base_column.value_type
            random_value = [
                cls.__DEFAULT_MAPPER[base_type]() for _ in range(length)
            ]
        elif column._meta.choices:
            random_value = RandomBuilder.next_enum(column._meta.choices)
        else:
            random_value = cls.__DEFAULT_MAPPER[column.value_type]()

        if "length" in column._meta.params and isinstance(random_value, str):
            return random_value[: column._meta.params["length"]]
        elif isinstance(column, (JSON, JSONB)):
            return json.dumps({"value": random_value})

        return random_value



from __future__ import annotations

import inspect
import typing as t
from dataclasses import dataclass

from piccolo.columns.column_types import (
    JSON,
    JSONB,
    Column,
    ForeignKey,
    LazyTableReference,
)
from piccolo.querystring import QueryString, Selectable
from piccolo.utils.list import flatten
from piccolo.utils.sync import run_sync

if t.TYPE_CHECKING:
    from piccolo.table import Table


class M2MSelect(Selectable):
    """
    This is a subquery used within a select to fetch data via an M2M table.
    """

    def __init__(
        self,
        *columns: Column,
        m2m: M2M,
        as_list: bool = False,
        load_json: bool = False,
    ):
        """
        :param columns:
            Which columns to include from the related table.
        :param as_list:
            If a single column is provided, and ``as_list`` is ``True`` a
            flattened list will be returned, rather than a list of objects.
        :param load_json:
            If ``True``, any JSON strings are loaded as Python objects.

        """
        self.as_list = as_list
        self.columns = columns
        self.m2m = m2m
        self.load_json = load_json

        safe_types = (int, str)

        # If the columns can be serialised / deserialised as JSON, then we
        # can fetch the data all in one go.
        self.serialisation_safe = all(
            (column.__class__.value_type in safe_types)
            and (type(column) not in (JSON, JSONB))
            for column in columns
        )

    def get_select_string(
        self, engine_type: str, with_alias=True
    ) -> QueryString:
        m2m_table_name_with_schema = (
            self.m2m._meta.resolved_joining_table._meta.get_formatted_tablename()  # noqa: E501
        )  # noqa: E501
        m2m_relationship_name = self.m2m._meta.name

        fk_1 = self.m2m._meta.primary_foreign_key
        fk_1_name = fk_1._meta.db_column_name
        table_1 = fk_1._foreign_key_meta.resolved_references
        table_1_name = table_1._meta.tablename
        table_1_name_with_schema = table_1._meta.get_formatted_tablename()
        table_1_pk_name = table_1._meta.primary_key._meta.db_column_name

        fk_2 = self.m2m._meta.secondary_foreign_key
        fk_2_name = fk_2._meta.db_column_name
        table_2 = fk_2._foreign_key_meta.resolved_references
        table_2_name = table_2._meta.tablename
        table_2_name_with_schema = table_2._meta.get_formatted_tablename()
        table_2_pk_name = table_2._meta.primary_key._meta.db_column_name

        inner_select = f"""
            {m2m_table_name_with_schema}
            JOIN {table_1_name_with_schema} "inner_{table_1_name}" ON (
                {m2m_table_name_with_schema}."{fk_1_name}" = "inner_{table_1_name}"."{table_1_pk_name}"
            )
            JOIN {table_2_name_with_schema} "inner_{table_2_name}" ON (
                {m2m_table_name_with_schema}."{fk_2_name}" = "inner_{table_2_name}"."{table_2_pk_name}"
            )
            WHERE {m2m_table_name_with_schema}."{fk_1_name}" = "{table_1_name}"."{table_1_pk_name}"
        """  # noqa: E501

        if engine_type in ("postgres", "cockroach"):
            if self.as_list:
                column_name = self.columns[0]._meta.db_column_name
                return QueryString(
                    f"""
                    ARRAY(
                        SELECT
                            "inner_{table_2_name}"."{column_name}"
                        FROM {inner_select}
                    ) AS "{m2m_relationship_name}"
                """
                )
            elif not self.serialisation_safe:
                column_name = table_2_pk_name
                return QueryString(
                    f"""
                    ARRAY(
                        SELECT
                            "inner_{table_2_name}"."{column_name}"
                        FROM {inner_select}
                    ) AS "{m2m_relationship_name}"
                """
                )
            else:
                column_names = ", ".join(
                    f'"inner_{table_2_name}"."{column._meta.db_column_name}"'
                    for column in self.columns
                )
                return QueryString(
                    f"""
                    (
                        SELECT JSON_AGG({m2m_relationship_name}_results)
                        FROM (
                            SELECT {column_names} FROM {inner_select}
                        ) AS "{m2m_relationship_name}_results"
                    ) AS "{m2m_relationship_name}"
                """
                )
        elif engine_type == "sqlite":
            if len(self.columns) > 1 or not self.serialisation_safe:
                column_name = table_2_pk_name
            else:
                assert len(self.columns) > 0
                column_name = self.columns[0]._meta.db_column_name

            return QueryString(
                f"""
                (
                    SELECT group_concat(
                        "inner_{table_2_name}"."{column_name}"
                    )
                    FROM {inner_select}
                )
                AS "{m2m_relationship_name} [M2M]"
            """
            )
        else:
            raise ValueError(f"{engine_type} is an unrecognised engine type")


@dataclass
class M2MMeta:
    joining_table: t.Union[t.Type[Table], LazyTableReference]
    _foreign_key_columns: t.Optional[t.List[ForeignKey]] = None

    # Set by the Table Metaclass:
    _name: t.Optional[str] = None
    _table: t.Optional[t.Type[Table]] = None

    @property
    def name(self) -> str:
        if not self._name:
            raise ValueError(
                "`_name` isn't defined - the Table Metaclass should set it."
            )
        return self._name

    @property
    def table(self) -> t.Type[Table]:
        if not self._table:
            raise ValueError(
                "`_table` isn't defined - the Table Metaclass should set it."
            )
        return self._table

    @property
    def resolved_joining_table(self) -> t.Type[Table]:
        """
        Evaluates the ``joining_table`` attribute if it's a
        ``LazyTableReference``, raising a ``ValueError`` if it fails, otherwise
        returns a ``Table`` subclass.
        """
        from piccolo.table import Table

        if isinstance(self.joining_table, LazyTableReference):
            return self.joining_table.resolve()
        elif inspect.isclass(self.joining_table) and issubclass(
            self.joining_table, Table
        ):
            return self.joining_table
        else:
            raise ValueError(
                "The joining_table attribute is neither a Table subclass or a "
                "LazyTableReference instance."
            )

    @property
    def foreign_key_columns(self) -> t.List[ForeignKey]:
        if not self._foreign_key_columns:
            self._foreign_key_columns = (
                self.resolved_joining_table._meta.foreign_key_columns[:2]
            )
        return self._foreign_key_columns

    @property
    def primary_foreign_key(self) -> ForeignKey:
        """
        The joining table has two foreign keys. We need a way to distinguish
        between them. The primary is the one which points to the table with
        ``M2M`` defined on it. In this example the primary foreign key is the
        one which points to ``Band``:

        .. code-block:: python

            class Band(Table):
                name = Varchar()
                genres = M2M(
                    LazyTableReference("GenreToBand", module_path=__name__)
                )

            class Genre(Table):
                name = Varchar()

            class GenreToBand(Table):
                band = ForeignKey(Band)  # primary
                genre = ForeignKey(Genre)  # secondary

        The secondary foreign key is the one which points to ``Genre``.

        """
        for fk_column in self.foreign_key_columns:
            if fk_column._foreign_key_meta.resolved_references == self.table:
                return fk_column

        raise ValueError("No matching foreign key column found!")

    @property
    def primary_table(self) -> t.Type[Table]:
        return self.primary_foreign_key._foreign_key_meta.resolved_references

    @property
    def secondary_foreign_key(self) -> ForeignKey:
        """
        See ``primary_foreign_key``.
        """
        for fk_column in self.foreign_key_columns:
            if fk_column._foreign_key_meta.resolved_references != self.table:
                return fk_column

        raise ValueError("No matching foreign key column found!")

    @property
    def secondary_table(self) -> t.Type[Table]:
        return self.secondary_foreign_key._foreign_key_meta.resolved_references


@dataclass
class M2MAddRelated:
    target_row: Table
    m2m: M2M
    rows: t.Sequence[Table]
    extra_column_values: t.Dict[t.Union[Column, str], t.Any]

    @property
    def resolved_extra_column_values(self) -> t.Dict[str, t.Any]:
        return {
            i._meta.name if isinstance(i, Column) else i: j
            for i, j in self.extra_column_values.items()
        }

    async def _run(self):
        rows = self.rows
        unsaved = [i for i in rows if not i._exists_in_db]

        if unsaved:
            await rows[0].__class__.insert(*unsaved).run()

        joining_table = self.m2m._meta.resolved_joining_table

        joining_table_rows = []

        for row in rows:
            joining_table_row = joining_table(
                **self.resolved_extra_column_values
            )
            setattr(
                joining_table_row,
                self.m2m._meta.primary_foreign_key._meta.name,
                getattr(
                    self.target_row,
                    self.target_row._meta.primary_key._meta.name,
                ),
            )
            setattr(
                joining_table_row,
                self.m2m._meta.secondary_foreign_key._meta.name,
                getattr(
                    row,
                    row._meta.primary_key._meta.name,
                ),
            )
            joining_table_rows.append(joining_table_row)

        return await joining_table.insert(*joining_table_rows).run()

    async def run(self):
        """
        Run the queries, making sure they are either within an existing
        transaction, or wrapped in a new transaction.
        """
        engine = self.rows[0]._meta.db
        if engine.transaction_exists():
            await self._run()
        else:
            async with engine.transaction():
                await self._run()

    def run_sync(self):
        return run_sync(self.run())

    def __await__(self):
        return self.run().__await__()


@dataclass
class M2MRemoveRelated:
    target_row: Table
    m2m: M2M
    rows: t.Sequence[Table]

    async def run(self):
        fk = self.m2m._meta.secondary_foreign_key
        related_table = fk._foreign_key_meta.resolved_references

        row_ids = []

        for row in self.rows:
            if row.__class__ != related_table:
                raise ValueError("The row belongs to the wrong table!")

            row_id = getattr(row, row._meta.primary_key._meta.name)
            if row_id:
                row_ids.append(row_id)

        if row_ids:
            return (
                await self.m2m._meta.resolved_joining_table.delete()
                .where(
                    self.m2m._meta.primary_foreign_key == self.target_row,
                    self.m2m._meta.secondary_foreign_key.is_in(row_ids),
                )
                .run()
            )

        return None

    def run_sync(self):
        return run_sync(self.run())

    def __await__(self):
        return self.run().__await__()


@dataclass
class M2MGetRelated:
    row: Table
    m2m: M2M

    async def run(self):
        joining_table = self.m2m._meta.resolved_joining_table

        secondary_table = self.m2m._meta.secondary_table

        # TODO - replace this with a subquery in the future.
        ids = (
            await joining_table.select(
                getattr(
                    self.m2m._meta.secondary_foreign_key,
                    secondary_table._meta.primary_key._meta.name,
                )
            )
            .where(self.m2m._meta.primary_foreign_key == self.row)
            .output(as_list=True)
        )

        results = await secondary_table.objects().where(
            secondary_table._meta.primary_key.is_in(ids)
        )

        return results

    def run_sync(self):
        return run_sync(self.run())

    def __await__(self):
        return self.run().__await__()


class M2M:
    def __init__(
        self,
        joining_table: t.Union[t.Type[Table], LazyTableReference],
        foreign_key_columns: t.Optional[t.List[ForeignKey]] = None,
    ):
        """
        :param joining_table:
            A ``Table`` containing two ``ForeignKey`` columns.
        :param foreign_key_columns:
            If for some reason your joining table has more than two foreign key
            columns, you can explicitly specify which two are relevant.

        """
        if foreign_key_columns and (
            len(foreign_key_columns) != 2
            or not all(isinstance(i, ForeignKey) for i in foreign_key_columns)
        ):
            raise ValueError("You must specify two ForeignKey columns.")

        self._meta = M2MMeta(
            joining_table=joining_table,
            _foreign_key_columns=foreign_key_columns,
        )

    def __call__(
        self,
        *columns: t.Union[Column, t.List[Column]],
        as_list: bool = False,
        load_json: bool = False,
    ) -> M2MSelect:
        """
        :param columns:
            Which columns to include from the related table. If none are
            specified, then all of the columns are returned.
        :param as_list:
            If a single column is provided, and ``as_list`` is ``True`` a
            flattened list will be returned, rather than a list of objects.
        :param load_json:
            If ``True``, any JSON strings are loaded as Python objects.
        """
        columns_ = flatten(columns)

        if not columns_:
            columns_ = self._meta.secondary_table._meta.columns

        if as_list and len(columns_) != 1:
            raise ValueError(
                "`as_list` is only valid with a single column argument"
            )

        return M2MSelect(
            *columns_, m2m=self, as_list=as_list, load_json=load_json
        )



from decimal import Decimal

from piccolo.columns.column_types import Numeric
from piccolo.table import Table
from tests.base import TableTest


class MyTable(Table):
    column_a = Numeric()
    column_b = Numeric(digits=(3, 2))


class TestNumeric(TableTest):
    tables = [MyTable]

    def test_creation(self):
        row = MyTable(column_a=Decimal(1.23), column_b=Decimal(1.23))
        row.save().run_sync()

        _row = MyTable.objects().first().run_sync()
        assert _row is not None

        self.assertEqual(type(_row.column_a), Decimal)
        self.assertEqual(type(_row.column_b), Decimal)

        self.assertAlmostEqual(_row.column_a, Decimal(1.23))
        self.assertAlmostEqual(_row.column_b, Decimal("1.23"))



from __future__ import annotations

import typing as t

try:
    import orjson

    ORJSON = True
except ImportError:
    import json

    ORJSON = False


def dump_json(data: t.Any, pretty: bool = False) -> str:
    if ORJSON:
        orjson_params: t.Dict[str, t.Any] = {"default": str}
        if pretty:
            orjson_params["option"] = (
                orjson.OPT_INDENT_2 | orjson.OPT_APPEND_NEWLINE  # type: ignore
            )
        return orjson.dumps(data, **orjson_params).decode(  # type: ignore
            "utf8"
        )
    else:
        params: t.Dict[str, t.Any] = {"default": str}
        if pretty:
            params["indent"] = 2
        return json.dumps(data, **params)  # type: ignore


class JSONDict(dict):
    """
    Once we have parsed a JSON string into a dictionary, we can't distinguish
    it from other dictionaries.

    Sometimes we might want to - for example::

        >>> await Album.select(
        ...     Album.all_columns(),
        ...     Album.recording_studio.all_columns()
        ... ).output(
        ...     nested=True,
        ...     load_json=True
        ... )

        [{
            'id': 1,
            'band': 1,
            'name': 'Awesome album 1',
            'recorded_at': {
                'id': 1,
                'facilities': {'restaurant': True, 'mixing_desk': True},
                'name': 'Abbey Road'
            },
            'release_date': datetime.date(2021, 1, 1)
        }]

    Facilities could be mistaken for a table.

    """

    ...


def load_json(data: str) -> t.Any:
    response = (
        orjson.loads(data) if ORJSON else json.loads(data)  # type: ignore
    )

    if isinstance(response, dict):
        return JSONDict(**response)

    return response



import unittest

from tests.example_apps.music.tables import Band


class TestWhere(unittest.TestCase):
    def test_equals(self):
        _where = Band.name == "Pythonistas"
        sql = _where.__str__()
        self.assertEqual(sql, '"band"."name" = \'Pythonistas\'')

    def test_not_equal(self):
        _where = Band.name != "CSharps"
        sql = _where.__str__()
        self.assertEqual(sql, '"band"."name" != \'CSharps\'')

    def test_like(self):
        _where = Band.name.like("Python%")
        sql = _where.__str__()
        self.assertEqual(sql, '"band"."name" LIKE \'Python%\'')

    def test_is_in(self):
        _where = Band.name.is_in(["Pythonistas", "Rustaceans"])
        sql = _where.__str__()
        self.assertEqual(
            sql, "\"band\".\"name\" IN ('Pythonistas', 'Rustaceans')"
        )

        with self.assertRaises(ValueError):
            Band.name.is_in([])

    def test_not_in(self):
        _where = Band.name.not_in(["CSharps"])
        sql = _where.__str__()
        self.assertEqual(sql, '"band"."name" NOT IN (\'CSharps\')')

        with self.assertRaises(ValueError):
            Band.name.not_in([])


class TestAnd(unittest.TestCase):
    def test_get_column_values(self):
        """
        Make sure that we can extract the column values from an ``And``.

        There was a bug with ``None`` values not working:

        https://github.com/piccolo-orm/piccolo/issues/715

        """
        And_ = (Band.manager.is_null()) & (Band.name == "Pythonistas")
        column_values = And_.get_column_values()

        self.assertDictEqual(
            column_values, {Band.name: "Pythonistas", Band.manager: None}
        )



from unittest import TestCase

from piccolo.columns import Serial, Varchar
from piccolo.table import Table


class Manager(Table):
    id: Serial
    name = Varchar(unique=True)
    email = Varchar(unique=True)


class Band(Table):
    id: Serial
    name = Varchar(unique=True)
    manager_name = Varchar()


class Concert(Table):
    id: Serial
    title = Varchar()
    band_name = Varchar()


class TestJoinOn(TestCase):
    tables = [Manager, Band, Concert]

    def setUp(self):
        for table in self.tables:
            table.create_table().run_sync()

        Manager.insert(
            Manager(name="Guido", email="guido@example.com"),
            Manager(name="Maz", email="maz@example.com"),
            Manager(name="Graydon", email="graydon@example.com"),
        ).run_sync()

        Band.insert(
            Band(name="Pythonistas", manager_name="Guido"),
            Band(name="Rustaceans", manager_name="Graydon"),
        ).run_sync()

        Concert.insert(
            Concert(
                title="Rockfest",
                band_name="Pythonistas",
            ),
        ).run_sync()

    def tearDown(self):
        for table in self.tables:
            table.alter().drop_table().run_sync()

    def test_join_on(self):
        """
        Do a simple join between two tables.
        """
        query = Band.select(
            Band.name,
            Band.manager_name,
            Band.manager_name.join_on(Manager.name).email.as_alias(
                "manager_email"
            ),
        ).order_by(Band.id)

        response = query.run_sync()

        self.assertListEqual(
            response,
            [
                {
                    "name": "Pythonistas",
                    "manager_name": "Guido",
                    "manager_email": "guido@example.com",
                },
                {
                    "name": "Rustaceans",
                    "manager_name": "Graydon",
                    "manager_email": "graydon@example.com",
                },
            ],
        )

    def test_deeper_join(self):
        """
        Do a join between three tables.
        """
        response = (
            Concert.select(
                Concert.title,
                Concert.band_name,
                Concert.band_name.join_on(Band.name)
                .manager_name.join_on(Manager.name)
                .email.as_alias("manager_email"),
            )
            .order_by(Concert.id)
            .run_sync()
        )

        self.assertListEqual(
            response,
            [
                {
                    "title": "Rockfest",
                    "band_name": "Pythonistas",
                    "manager_email": "guido@example.com",
                }
            ],
        )



import datetime

from piccolo.columns import Timestamp
from piccolo.query.functions.datetime import (
    Day,
    Extract,
    Hour,
    Minute,
    Month,
    Second,
    Strftime,
    Year,
)
from piccolo.table import Table
from tests.base import TableTest, engines_only, sqlite_only


class Concert(Table):
    starts = Timestamp()


class DatetimeTest(TableTest):
    tables = [Concert]

    def setUp(self) -> None:
        super().setUp()
        self.concert = Concert(
            {
                Concert.starts: datetime.datetime(
                    year=2024, month=6, day=14, hour=23, minute=46, second=10
                )
            }
        )
        self.concert.save().run_sync()


@engines_only("postgres", "cockroach")
class TestExtract(DatetimeTest):
    def test_extract(self):
        self.assertEqual(
            Concert.select(
                Extract(Concert.starts, "year", alias="starts_year")
            ).run_sync(),
            [{"starts_year": self.concert.starts.year}],
        )

    def test_invalid_format(self):
        with self.assertRaises(ValueError):
            Extract(
                Concert.starts,
                "abc123",  # type: ignore
                alias="starts_year",
            )


@sqlite_only
class TestStrftime(DatetimeTest):
    def test_strftime(self):
        self.assertEqual(
            Concert.select(
                Strftime(Concert.starts, "%Y", alias="starts_year")
            ).run_sync(),
            [{"starts_year": str(self.concert.starts.year)}],
        )


class TestDatabaseAgnostic(DatetimeTest):
    def test_year(self):
        self.assertEqual(
            Concert.select(
                Year(Concert.starts, alias="starts_year")
            ).run_sync(),
            [{"starts_year": self.concert.starts.year}],
        )

    def test_month(self):
        self.assertEqual(
            Concert.select(
                Month(Concert.starts, alias="starts_month")
            ).run_sync(),
            [{"starts_month": self.concert.starts.month}],
        )

    def test_day(self):
        self.assertEqual(
            Concert.select(Day(Concert.starts, alias="starts_day")).run_sync(),
            [{"starts_day": self.concert.starts.day}],
        )

    def test_hour(self):
        self.assertEqual(
            Concert.select(
                Hour(Concert.starts, alias="starts_hour")
            ).run_sync(),
            [{"starts_hour": self.concert.starts.hour}],
        )

    def test_minute(self):
        self.assertEqual(
            Concert.select(
                Minute(Concert.starts, alias="starts_minute")
            ).run_sync(),
            [{"starts_minute": self.concert.starts.minute}],
        )

    def test_second(self):
        self.assertEqual(
            Concert.select(
                Second(Concert.starts, alias="starts_second")
            ).run_sync(),
            [{"starts_second": self.concert.starts.second}],
        )



import datetime

from piccolo.columns.column_types import Date
from piccolo.columns.defaults.date import DateNow
from piccolo.table import Table
from tests.base import TableTest


class MyTable(Table):
    created_on = Date()


class MyTableDefault(Table):
    created_on = Date(default=DateNow())


class TestDate(TableTest):
    tables = [MyTable]

    def test_timestamp(self):
        created_on = datetime.datetime.now().date()
        row = MyTable(created_on=created_on)
        row.save().run_sync()

        result = MyTable.objects().first().run_sync()
        assert result is not None
        self.assertEqual(result.created_on, created_on)


class TestDateDefault(TableTest):
    tables = [MyTableDefault]

    def test_timestamp(self):
        created_on = datetime.datetime.now().date()
        row = MyTableDefault()
        row.save().run_sync()

        result = MyTableDefault.objects().first().run_sync()
        assert result is not None
        self.assertEqual(result.created_on, created_on)



from piccolo.columns.operators.base import Operator


class ComparisonOperator(Operator):
    template = ""


class IsNull(ComparisonOperator):
    template = "{name} IS NULL"


class IsNotNull(ComparisonOperator):
    template = "{name} IS NOT NULL"


class Equal(ComparisonOperator):
    template = "{name} = {value}"


class NotEqual(ComparisonOperator):
    template = "{name} != {value}"


class In(ComparisonOperator):
    template = "{name} IN ({values})"


class NotIn(ComparisonOperator):
    template = "{name} NOT IN ({values})"


class Like(ComparisonOperator):
    template = "{name} LIKE {value}"


class ILike(ComparisonOperator):
    template = "{name} ILIKE {value}"


class NotLike(ComparisonOperator):
    template = "{name} NOT LIKE {value}"


class GreaterThan(ComparisonOperator):
    # Add permitted types???
    template = "{name} > {value}"


class GreaterEqualThan(ComparisonOperator):
    template = "{name} >= {value}"


class LessThan(ComparisonOperator):
    template = "{name} < {value}"


class LessEqualThan(ComparisonOperator):
    template = "{name} <= {value}"


class ArrayAny(ComparisonOperator):
    template = "{value} = ANY ({name})"


class ArrayNotAny(ComparisonOperator):
    template = "NOT {value} = ANY ({name})"


class ArrayAll(ComparisonOperator):
    template = "{value} = ALL ({name})"



from piccolo.columns.column_types import JSONB, ForeignKey, Varchar
from piccolo.table import Table
from tests.base import TableTest, engines_only, engines_skip


class RecordingStudio(Table):
    name = Varchar()
    facilities = JSONB(null=True)


class Instrument(Table):
    name = Varchar()
    studio = ForeignKey(RecordingStudio)


@engines_only("postgres", "cockroach")
class TestJSONB(TableTest):
    tables = [RecordingStudio, Instrument]

    def test_json(self):
        """
        Test storing a valid JSON string.
        """
        row = RecordingStudio(
            name="Abbey Road", facilities='{"mixing_desk": true}'
        )
        row.save().run_sync()
        self.assertEqual(row.facilities, '{"mixing_desk": true}')

    @engines_skip("cockroach")
    def test_raw(self):
        """
        Make sure raw queries convert the Python value into a JSON string.
        """
        RecordingStudio.raw(
            "INSERT INTO recording_studio (name, facilities) VALUES ({}, {})",
            "Abbey Road",
            '{"mixing_desk": true}',
        ).run_sync()

        self.assertEqual(
            RecordingStudio.select().run_sync(),
            [
                {
                    "id": 1,
                    "name": "Abbey Road",
                    "facilities": '{"mixing_desk": true}',
                }
            ],
        )

    @engines_only("cockroach")
    def test_raw_alt(self):
        """
        Make sure raw queries convert the Python value into a JSON string.
        """
        result = RecordingStudio.raw(
            "INSERT INTO recording_studio (name, facilities) VALUES ({}, {}) returning id",  # noqa: E501
            "Abbey Road",
            '{"mixing_desk": true}',
        ).run_sync()

        self.assertEqual(
            RecordingStudio.select().run_sync(),
            [
                {
                    "id": result[0]["id"],
                    "name": "Abbey Road",
                    "facilities": '{"mixing_desk": true}',
                }
            ],
        )

    def test_where(self):
        """
        Test using the where clause to match a subset of rows.
        """
        RecordingStudio.insert(
            RecordingStudio(
                name="Abbey Road", facilities={"mixing_desk": True}
            ),
            RecordingStudio(name="ABC Studio", facilities=None),
        ).run_sync()

        self.assertEqual(
            RecordingStudio.select(RecordingStudio.name)
            .where(RecordingStudio.facilities == {"mixing_desk": True})
            .run_sync(),
            [{"name": "Abbey Road"}],
        )

        self.assertEqual(
            RecordingStudio.select(RecordingStudio.name)
            .where(RecordingStudio.facilities == '{"mixing_desk": true}')
            .run_sync(),
            [{"name": "Abbey Road"}],
        )

        self.assertEqual(
            RecordingStudio.select(RecordingStudio.name)
            .where(RecordingStudio.facilities.is_null())
            .run_sync(),
            [{"name": "ABC Studio"}],
        )

        self.assertEqual(
            RecordingStudio.select(RecordingStudio.name)
            .where(RecordingStudio.facilities.is_not_null())
            .run_sync(),
            [{"name": "Abbey Road"}],
        )

    def test_as_alias_join(self):
        """
        Make sure that ``as_alias`` performs correctly when used via a join.
        """
        studio = (
            RecordingStudio.objects()
            .create(name="Abbey Road", facilities={"mixing_desk": True})
            .run_sync()
        )

        Instrument.objects().create(name="Guitar", studio=studio).run_sync()

        response = (
            Instrument.select(
                Instrument.name,
                Instrument.studio.facilities.as_alias("studio_facilities"),
            )
            .output(load_json=True)
            .run_sync()
        )

        self.assertListEqual(
            response,
            [{"name": "Guitar", "studio_facilities": {"mixing_desk": True}}],
        )

    def test_arrow(self):
        """
        Test using the arrow function to retrieve a subset of the JSON.
        """
        RecordingStudio(
            name="Abbey Road", facilities='{"mixing_desk": true}'
        ).save().run_sync()

        row = (
            RecordingStudio.select(
                RecordingStudio.facilities.arrow("mixing_desk")
            )
            .first()
            .run_sync()
        )
        assert row is not None
        self.assertEqual(row["facilities"], "true")

        row = (
            RecordingStudio.select(
                RecordingStudio.facilities.arrow("mixing_desk")
            )
            .output(load_json=True)
            .first()
            .run_sync()
        )
        assert row is not None
        self.assertEqual(row["facilities"], True)

    def test_arrow_as_alias(self):
        """
        Test using the arrow function to retrieve a subset of the JSON.
        """
        RecordingStudio(
            name="Abbey Road", facilities='{"mixing_desk": true}'
        ).save().run_sync()

        row = (
            RecordingStudio.select(
                RecordingStudio.facilities.arrow("mixing_desk").as_alias(
                    "mixing_desk"
                )
            )
            .first()
            .run_sync()
        )
        assert row is not None
        self.assertEqual(row["mixing_desk"], "true")

    def test_arrow_where(self):
        """
        Make sure the arrow function can be used within a WHERE clause.
        """
        RecordingStudio(
            name="Abbey Road", facilities='{"mixing_desk": true}'
        ).save().run_sync()

        self.assertEqual(
            RecordingStudio.count()
            .where(RecordingStudio.facilities.arrow("mixing_desk").eq(True))
            .run_sync(),
            1,
        )

        self.assertEqual(
            RecordingStudio.count()
            .where(RecordingStudio.facilities.arrow("mixing_desk").eq(False))
            .run_sync(),
            0,
        )

    def test_arrow_first(self):
        """
        Make sure the arrow function can be used with the first clause.
        """
        RecordingStudio.insert(
            RecordingStudio(facilities='{"mixing_desk": true}'),
            RecordingStudio(facilities='{"mixing_desk": false}'),
        ).run_sync()

        self.assertEqual(
            RecordingStudio.select(
                RecordingStudio.facilities.arrow("mixing_desk").as_alias(
                    "mixing_desk"
                )
            )
            .first()
            .run_sync(),
            {"mixing_desk": "true"},
        )



from typing import List


def get_fixed_length_string(string: str, length: int = 20) -> str:
    """
    Add spacing to the end of the string so it's a fixed length, or truncate
    if it's too long.
    """
    if len(string) > length:
        return f"{string[: length - 3]}..."
    spacing = "".join(" " for _ in range(length - len(string)))
    return f"{string}{spacing}"


def print_heading(string: str, width: int = 64) -> None:
    """
    Prints out a nicely formatted heading to the console. Useful for breaking
    up the output in large CLI commands.
    """
    print(f"\n{string.upper():^{width}}")
    print("-" * width)


def print_dict_table(data: List[dict], header_separator: bool = False) -> None:
    """
    Prints out a list of dictionaries in tabular form.

    Uses the first list element to extract the column names and their order
    within the row.

    """
    if len(data) == 0:
        raise ValueError("The data must have at least one element.")

    column_names = data[0].keys()
    widths = {column_name: len(column_name) for column_name in column_names}

    for item in data:
        for column in column_names:
            width = len(str(item[column]))
            if width > widths[column]:
                widths[column] = width

    format_string = " | ".join(f"{{:<{widths[w]}}}" for w in column_names)

    print(format_string.format(*[str(w) for w in column_names]))

    if header_separator:
        format_string_sep = "-+-".join(
            [f"{{:<{widths[w]}}}" for w in column_names]
        )
        print(
            format_string_sep.format(*["-" * widths[w] for w in column_names])
        )

    for item in data:
        print(format_string.format(*[str(item[w]) for w in column_names]))



from unittest import TestCase

from piccolo.columns import Integer, Varchar
from piccolo.table import Table


class Band(Table):
    name = Varchar(default=None, null=False)
    popularity = Integer()


class TestCreate(TestCase):
    def setUp(self):
        Band.create_table().run_sync()

    def tearDown(self):
        Band.alter().drop_table().run_sync()

    def test_create_new(self):
        """
        Make sure that creating a new instance works.
        """
        Band.objects().create(name="Pythonistas", popularity=1000).run_sync()

        names = [i["name"] for i in Band.select(Band.name).run_sync()]
        self.assertTrue("Pythonistas" in names)

    def test_null_values(self):
        """
        Make sure we test non-null columns:
        https://github.com/piccolo-orm/piccolo/issues/652
        """
        with self.assertRaises(ValueError) as manager:
            Band.objects().create().run_sync()

        self.assertEqual(str(manager.exception), "name wasn't provided")

        # Shouldn't raise an exception
        Band.objects().create(name="Pythonistas").run_sync()



from unittest import TestCase

from piccolo.apps.migrations.auto import MigrationManager, SchemaSnapshot


class TestSchemaSnaphot(TestCase):
    def test_add_table(self):
        """
        Test adding tables.
        """
        manager_1 = MigrationManager()
        manager_1.add_table(class_name="Manager", tablename="manager")

        manager_2 = MigrationManager()
        manager_2.add_table(class_name="Band", tablename="band")

        schema_snapshot = SchemaSnapshot(managers=[manager_1, manager_2])
        snapshot = schema_snapshot.get_snapshot()

        self.assertTrue(len(snapshot) == 2)

        class_names = [i.class_name for i in snapshot]
        self.assertTrue("Band" in class_names)
        self.assertTrue("Manager" in class_names)

    def test_drop_table(self):
        """
        Test dropping tables.
        """
        manager_1 = MigrationManager()
        manager_1.add_table(class_name="Manager", tablename="manager")
        manager_1.add_table(class_name="Band", tablename="band")

        manager_2 = MigrationManager()
        manager_2.drop_table(class_name="Band", tablename="band")

        schema_snapshot = SchemaSnapshot(managers=[manager_1, manager_2])
        snapshot = schema_snapshot.get_snapshot()

        self.assertTrue(len(snapshot) == 1)

        class_names = [i.class_name for i in snapshot]
        self.assertTrue("Manager" in class_names)

    def test_rename_table(self):
        """
        Test renaming tables.
        """
        manager_1 = MigrationManager()
        manager_1.add_table(class_name="Band", tablename="band")

        manager_2 = MigrationManager()
        manager_2.rename_table(
            old_class_name="Band",
            old_tablename="band",
            new_class_name="Performer",
            new_tablename="performer",
        )

        schema_snapshot = SchemaSnapshot(managers=[manager_1, manager_2])
        snapshot = schema_snapshot.get_snapshot()

        self.assertTrue(snapshot[0].class_name == "Performer")
        self.assertTrue(snapshot[0].tablename == "performer")

    def test_add_column(self):
        """
        Test adding columns.
        """
        manager = MigrationManager()
        manager.add_table(class_name="Manager", tablename="manager")
        manager.add_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            column_class_name="Varchar",
            params={"length": 100},
        )

        schema_snapshot = SchemaSnapshot(managers=[manager])
        snapshot = schema_snapshot.get_snapshot()

        self.assertTrue(len(snapshot) == 1)
        self.assertTrue(len(snapshot[0].columns) == 1)

    def test_rename_column(self):
        """
        Test renaming columns.
        """
        manager_1 = MigrationManager()
        manager_1.add_table(class_name="Manager", tablename="manager")
        manager_1.add_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            column_class_name="Varchar",
            params={"length": 100},
        )

        manager_2 = MigrationManager()
        manager_2.rename_column(
            table_class_name="Manager",
            tablename="manager",
            old_column_name="name",
            new_column_name="title",
        )

        schema_snapshot = SchemaSnapshot(managers=[manager_1, manager_2])
        snapshot = schema_snapshot.get_snapshot()
        self.assertTrue(snapshot[0].columns[0]._meta.name == "title")

        # Make sure double renames still work
        manager_3 = MigrationManager()
        manager_3.rename_column(
            table_class_name="Manager",
            tablename="manager",
            old_column_name="title",
            new_column_name="label",
        )

        schema_snapshot = SchemaSnapshot(
            managers=[manager_1, manager_2, manager_3]
        )
        snapshot = schema_snapshot.get_snapshot()
        self.assertTrue(snapshot[0].columns[0]._meta.name == "label")

    def test_drop_column(self):
        """
        Test dropping columns.
        """
        manager_1 = MigrationManager()
        manager_1.add_table(class_name="Manager", tablename="manager")
        manager_1.add_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            column_class_name="Varchar",
            params={"length": 100},
        )

        manager_2 = MigrationManager()
        manager_2.drop_column(
            table_class_name="Manager", tablename="manager", column_name="name"
        )

        schema_snapshot = SchemaSnapshot(managers=[manager_1, manager_2])
        snapshot = schema_snapshot.get_snapshot()
        self.assertEqual(len(snapshot[0].columns), 0)

    def test_alter_column(self):
        """
        Test altering columns.
        """
        manager_1 = MigrationManager()
        manager_1.add_table(class_name="Manager", tablename="manager")
        manager_1.add_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            column_class_name="Varchar",
            params={"length": 100},
        )

        manager_2 = MigrationManager()
        manager_2.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"unique": True},
            old_params={"unique": False},
        )

        schema_snapshot = SchemaSnapshot(managers=[manager_1, manager_2])
        snapshot = schema_snapshot.get_snapshot()

        self.assertTrue(snapshot[0].columns[0]._meta.unique)

    def test_get_table_from_snapshot(self):
        manager_1 = MigrationManager()
        manager_1.add_table(class_name="Manager", tablename="manager")
        manager_1.add_table(class_name="Band", tablename="band")

        schema_snapshot = SchemaSnapshot(managers=[manager_1])
        table = schema_snapshot.get_table_from_snapshot("Manager")

        self.assertTrue(table.class_name == "Manager")

        with self.assertRaises(ValueError):
            schema_snapshot.get_table_from_snapshot("Foo")
